from typing import List, Optional
from abc import abstractmethod
import numpy as np
import pyterrier as pt
import pandas as pd
import pyterrier_alpha as pta
from . import SimFn


class BiEncoder(pt.Transformer):
    """Represents a single-vector dense bi-encoder.

    A ``BiEncoder`` encodes the text of a query or document into a dense vector.

    This class functions as a transformer factory:
     - Query encoding using :meth:`query_encoder`
     - Document encoding using :meth:`doc_encoder`
     - Text scoring (re-reranking) using :meth:`text_scorer`

    It can also be used as a transformer directly. It infers which transformer to use
    based on columns present in the input frame.

    Note that in most cases, you will want to use a ``BiEncoder`` as part of a pipeline
    with a :class:`~pyterrier_dr.FlexIndex` to perform dense indexing and retrival.
    """
    def __init__(self, *, batch_size=32, text_field='text', verbose=False):
        """
        Args:
            batch_size: The default batch size to use for query/document encoding
            text_field: The field in the input dataframe that contains the document text
            verbose: Whether to show progress bars
        """
        super().__init__()
        self.batch_size = batch_size
        self.text_field = text_field
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        with pta.validate.any(inp) as v:
            v.columns(includes=['query', self.text_field], mode='scorer')
            v.columns(includes=['query_vec', self.text_field], mode='scorer')
            v.columns(includes=['query', 'doc_vec'], mode='scorer')
            v.columns(includes=['query_vec', 'doc_vec'], mode='scorer')
            v.columns(includes=['query'], mode='query_encoder')
            v.columns(includes=[self.text_field], mode='doc_encoder')

        if v.mode == 'scorer':
            return self.scorer()(inp)
        elif v.mode == 'query_encoder':
            return self.query_encoder()(inp)
        elif v.mode == 'doc_encoder':
            return self.doc_encoder()(inp)

    def query_encoder(self, verbose=None, batch_size=None) -> pt.Transformer:
        """
        Query encoding
        """
        return BiQueryEncoder(self, verbose=verbose, batch_size=batch_size)

    def doc_encoder(self, verbose=None, batch_size=None) -> pt.Transformer:
        """
        Doc encoding
        """
        return BiDocEncoder(self, verbose=verbose, batch_size=batch_size)

    def text_scorer(self, verbose=None, batch_size=None, sim_fn=None) -> pt.Transformer:
        """
        Text Scoring (re-ranking)
        """
        return BiScorer(self, verbose=verbose, batch_size=batch_size, sim_fn=sim_fn)

    def scorer(self, verbose=None, batch_size=None, sim_fn=None) -> pt.Transformer:
        return self.text_scorer(verbose=verbose, batch_size=batch_size, sim_fn=sim_fn)

    @property
    def sim_fn(self) -> SimFn:
        """
        The similarity function to use between embeddings for this model
        """
        if hasattr(self, 'config') and hasattr(self.config, 'sim_fn'):
            return SimFn(self.config.sim_fn)
        return SimFn.dot # default

    @abstractmethod
    def encode_queries(self, texts: List[str], batch_size: Optional[int] = None) -> np.array:
        """Abstract method to encode a list of query texts into dense vectors.

        This function is used by the transformer returned by :meth:`query_encoder`.

        Args:
            texts: A list of query texts
            batch_size: The batch size to use for encoding

        Returns:
            np.array: A numpy array of shape (n_queries, n_dims)
        """
        raise NotImplementedError()

    @abstractmethod
    def encode_docs(self, texts: List[str], batch_size: Optional[int] = None) -> np.array:
        """Abstract method to encode a list of document texts into dense vectors.

        This function is used by the transformer returned by :meth:`doc_encoder`.

        Args:
            texts: A list of document texts
            batch_size: The batch size to use for encoding

        Returns:
            np.array: A numpy array of shape (n_docs, n_dims)
        """
        raise NotImplementedError()


class BiQueryEncoder(pt.Transformer):
    def __init__(self, bi_encoder_model: BiEncoder, verbose=None, batch_size=None):
        self.bi_encoder_model = bi_encoder_model
        self.verbose = verbose if verbose is not None else bi_encoder_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_encoder_model.batch_size

    def encode(self, texts, batch_size=None) -> np.array:
        return self.bi_encoder_model.encode_queries(texts, batch_size=batch_size or self.batch_size)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=['query'])
        it = inp['query'].values
        it, inv = np.unique(it, return_inverse=True)
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Queries', unit='query')
        enc = self.encode(it)
        return inp.assign(query_vec=[enc[i] for i in inv])

    def __repr__(self):
        return f'{repr(self.bi_encoder_model)}.query_encoder()'


class BiDocEncoder(pt.Transformer):
    def __init__(self, bi_encoder_model: BiEncoder, verbose=None, batch_size=None, text_field=None):
        self.bi_encoder_model = bi_encoder_model
        self.verbose = verbose if verbose is not None else bi_encoder_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_encoder_model.batch_size
        self.text_field = text_field if text_field is not None else bi_encoder_model.text_field

    def encode(self, texts, batch_size=None) -> np.array:
        return self.bi_encoder_model.encode_docs(texts, batch_size=batch_size or self.batch_size)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=[self.text_field])
        it = inp[self.text_field]
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Docs', unit='doc')
        return inp.assign(doc_vec=list(self.encode(it)))

    def __repr__(self):
        return f'{repr(self.bi_encoder_model)}.doc_encoder()'


class BiScorer(pt.Transformer):
    def __init__(self, bi_encoder_model: BiEncoder, verbose=None, batch_size=None, text_field=None, sim_fn=None):
        self.bi_encoder_model = bi_encoder_model
        self.verbose = verbose if verbose is not None else bi_encoder_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_encoder_model.batch_size
        self.text_field = text_field if text_field is not None else bi_encoder_model.text_field
        self.sim_fn = sim_fn if sim_fn is not None else bi_encoder_model.sim_fn

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        with pta.validate.any(inp) as v:
            v.columns(includes=['query_vec', 'doc_vec'])
            v.columns(includes=['query', 'doc_vec'])
            v.columns(includes=['query_vec', self.text_field])
            v.columns(includes=['query', self.text_field])
        if 'query_vec' in inp.columns:
            query_vec = inp['query_vec']
        else:
            query_vec = self.bi_encoder_model.query_encoder(batch_size=self.batch_size, verbose=self.verbose)(inp)['query_vec']
        if 'doc_vec' in inp.columns:
            doc_vec = inp['doc_vec']
        else:
            doc_vec = self.bi_encoder_model.doc_encoder(batch_size=self.batch_size, verbose=self.verbose)(inp)['doc_vec']
        if self.sim_fn == SimFn.dot:
            scores = (query_vec * doc_vec).apply(np.sum)
        else:
            raise ValueError(f'{self.sim_fn} not yet supported by BiScorer')
        outp = inp.assign(score=scores)
        return pt.model.add_ranks(outp)

    def __repr__(self):
        return f'{repr(self.bi_encoder_model)}.scorer()'



import unittest
import numpy as np
import pandas as pd
from pyterrier_dr import AveragePrf, VectorPrf


class TestModels(unittest.TestCase):

    def test_average_prf(self):
        prf = AveragePrf()
        with self.subTest('single row'):
            inp = pd.DataFrame([['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])]], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            self.assertEqual(out['qid'].iloc[0], 'q1')
            self.assertEqual(out['query'].iloc[0], 'query')
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([2.5, 3.5, 4.5]))

        with self.subTest('multiple rows'):
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([3.5, 4.5, 3.]))

        with self.subTest('multiple rows -- k=3'):
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
                ['q1', 'query', np.array([1, 2, 3]), 'd4', np.array([100, 100, 100])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([3.5, 4.5, 3.]))

        with self.subTest('multiple rows -- k=1'):
            prf.k = 1
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
                ['q1', 'query', np.array([1, 2, 3]), 'd4', np.array([100, 100, 100])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([2.5, 3.5, 4.5]))

        with self.subTest('multiple queries'):
            prf.k = 3
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([8, 7, 1])],
                ['q2', 'query2', np.array([4, 6, 1]), 'd1', np.array([9, 4, 2])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 2)
            self.assertEqual(out['qid'].iloc[0], 'q1')
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([3.5, 4.5, 3.]))
            self.assertEqual(out['qid'].iloc[1], 'q2')
            np.testing.assert_array_equal(out['query_vec'].iloc[1], np.array([6.5, 5., 1.5]))

    def test_vector_prf(self):
        prf = VectorPrf(alpha=0.5, beta=0.5)
        with self.subTest('single row'):
            inp = pd.DataFrame([['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])]], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            self.assertEqual(out['qid'].iloc[0], 'q1')
            self.assertEqual(out['query'].iloc[0], 'query')
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([2.5, 3.5, 4.5]))

        with self.subTest('multiple rows'):
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_almost_equal(out['query_vec'].iloc[0], np.array([2.666667, 3.666667, 3.]), decimal=5)

        with self.subTest('multiple rows -- k=3'):
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
                ['q1', 'query', np.array([1, 2, 3]), 'd4', np.array([100, 100, 100])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_almost_equal(out['query_vec'].iloc[0], np.array([2.666667, 3.666667, 3.]), decimal=5)

        with self.subTest('multiple rows -- k=1'):
            prf.k = 1
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd2', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd3', np.array([8, 7, 1])],
                ['q1', 'query', np.array([1, 2, 3]), 'd4', np.array([100, 100, 100])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 1)
            np.testing.assert_array_equal(out['query_vec'].iloc[0], np.array([2.5, 3.5, 4.5]))

        with self.subTest('multiple queries'):
            prf.k = 3
            inp = pd.DataFrame([
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([4, 5, 6])],
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([1, 4, 2])],
                ['q1', 'query', np.array([1, 2, 3]), 'd1', np.array([8, 7, 1])],
                ['q2', 'query2', np.array([4, 6, 1]), 'd1', np.array([9, 4, 2])],
            ], columns=['qid', 'query', 'query_vec', 'docno', 'doc_vec'])
            out = prf(inp)
            self.assertEqual(out.columns.tolist(), ['qid', 'query', 'query_vec'])
            self.assertEqual(len(out), 2)
            self.assertEqual(out['qid'].iloc[0], 'q1')
            np.testing.assert_almost_equal(out['query_vec'].iloc[0], np.array([2.666667, 3.666667, 3.]), decimal=5)
            self.assertEqual(out['qid'].iloc[1], 'q2')
            np.testing.assert_array_equal(out['query_vec'].iloc[1], np.array([6.5, 5., 1.5]))



if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta


class VectorPrf(pt.Transformer):
    """
    Performs a Rocchio-esque PRF by linearly combining the query_vec column with 
    the doc_vec column of the top k documents.

    Arguments:
     - alpha: weight of original query_vec
     - beta: weight of doc_vec
     - k: number of pseudo-relevant feedback documents

    Expected Input Columns: ``['qid', 'query_vec', 'docno', 'doc_vec']``

    Output Columns: ``['qid', 'query_vec']`` (Any other query columns from the input are also pulled included in the output.)

    Example::
    
            prf_pipe = model >> index >> index.vec_loader() >> pyterier_dr.vector_prf() >> index 

    .. cite.dblp:: journals/tois/0009MZKZ23
    """
    def __init__(self,
        *,
        alpha: float = 1,
        beta: float = 0.2,
        k: int = 3
    ):
        self.alpha = alpha
        self.beta = beta
        self.k = k

    @pta.transform.by_query(add_ranks=False)
    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Performs Vector PRF on the input dataframe."""
        pta.validate.result_frame(inp, extra_columns=['query_vec', 'doc_vec'])

        query_cols = [col for col in inp.columns if col.startswith('q') and col != 'query_vec']

        # get the docvectors for the top k docs
        doc_vecs = np.stack([ row.doc_vec for row in inp.head(self.k).itertuples() ])
        # combine their average and add to the query
        query_vec = self.alpha * inp['query_vec'].iloc[0] + self.beta * np.mean(doc_vecs, axis=0)
        # generate new query dataframe with the existing query columns and the new query_vec
        return pd.DataFrame([[inp[c].iloc[0] for c in query_cols] + [query_vec]], columns=query_cols + ['query_vec'])

    def __repr__(self):
        return f"VectorPrf(alpha={self.alpha}, beta={self.beta}, k={self.k})"


class AveragePrf(pt.Transformer):
    """
    Performs Average PRF (as described by Li et al.) by averaging the query_vec column with 
    the doc_vec column of the top k documents.

    Arguments:
     - k: number of pseudo-relevant feedback documents

    Expected Input Columns: ``['qid', 'query_vec', 'docno', 'doc_vec']``

    Output Columns: ``['qid', 'query_vec']`` (Any other query columns from the input are also pulled included in the output.)

    Example::
    
            prf_pipe = model >> index >> index.vec_loader() >> pyterier_dr.average_prf() >> index 

    .. cite.dblp:: journals/tois/0009MZKZ23
    """
    def __init__(self,
        *,
        k: int = 3
    ):
        self.k = k

    @pta.transform.by_query(add_ranks=False)
    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Performs Average PRF on the input dataframe."""
        pta.validate.result_frame(inp, extra_columns=['query_vec', 'doc_vec'])

        query_cols = [col for col in inp.columns if col.startswith('q') and col != 'query_vec']

        # get the docvectors for the top k docs and the query_vec
        all_vecs = np.stack([inp['query_vec'].iloc[0]] + [row.doc_vec for row in inp.head(self.k).itertuples()])
        # combine their average and add to the query
        query_vec = np.mean(all_vecs, axis=0)
        # generate new query dataframe with the existing query columns and the new query_vec
        return pd.DataFrame([[inp[c].iloc[0] for c in query_cols] + [query_vec]], columns=query_cols + ['query_vec'])

    def __repr__(self):
        return f"AveragePrf(k={self.k})"



from typing import List, Optional, Union
import json
import os
import random
import numpy as np
import torch
import pyterrier as pt
from transformers import AutoConfig
from .biencoder import BiEncoder
from tqdm import tqdm
import pyterrier_alpha as pta


class CDE(BiEncoder):
    def __init__(self, model_name='jxm/cde-small-v1', cache: Optional['CDECache'] = None, batch_size=32, text_field='text', verbose=False, device=None):
        super().__init__(batch_size=batch_size, text_field=text_field, verbose=verbose)
        self.model_name = model_name
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        from sentence_transformers import SentenceTransformer
        self.model = SentenceTransformer(model_name, trust_remote_code=True).to(self.device).eval()
        self.config = AutoConfig.from_pretrained(model_name, trust_remote_code=True)
        self.cache = cache or CDECache(cde=self)

    def encode_context(self, texts: List[str], batch_size=None):
        show_progress = False
        if isinstance(texts, tqdm):
            texts.disable = True
            show_progress = True
        texts = list(texts)
        if len(texts) == 0:
            return np.empty(shape=(0, 0))
        return self.model.encode(
            texts,
            prompt_name="document",
            batch_size=batch_size or self.batch_size,
            show_progress=show_progress,
        )

    def encode_queries(self, texts: List[str], batch_size=None):
        show_progress = False
        if isinstance(texts, tqdm):
            texts.disable = True
            show_progress = True
        texts = list(texts)
        if len(texts) == 0:
            return np.empty(shape=(0, 0))
        result = self.model.encode(
            texts,
            prompt_name='query',
            dataset_embeddings=self.cache.context(),
            show_progress=show_progress,
        )
        # sentence transformers doesn't norm?
        result = result / np.linalg.norm(result, ord=2, axis=1, keepdims=True)
        return result

    def encode_docs(self, texts: List[str], batch_size=None):
        show_progress = False
        if isinstance(texts, tqdm):
            texts.disable = True
            show_progress = True
        texts = list(texts)
        if len(texts) == 0:
            return np.empty(shape=(0, 0))
        result = self.model.encode(
            texts,
            prompt_name='document',
            dataset_embeddings=self.cache.context(),
            show_progress=show_progress,
        )
        # sentence transformers doesn't norm?
        result = result / np.linalg.norm(result, ord=2, axis=1, keepdims=True)
        return result

    def __repr__(self):
        return f'CDE({self.model_name!r})'


def _sample_iter(it, k, rng=None):
    if rng is None:
        rng = random.Random(42)
    elif not isinstance(rng, random.Random):
        rng = random.Random(rng)
    result = []
    for i, item in enumerate(it):
        if i < k:
            result.append(item)
        elif (e := rng.randint(0, i)) < k:
            result[e] = item
    return result


class CDECache(pta.Artifact, pt.Indexer):
    ARTIFACT_TYPE = 'cde_cache'
    ARTIFACT_FORMAT = 'np_pickle'

    def __init__(self, path: Optional[str] = None, cde: Optional[Union[CDE, str]] = None, rng = None):
        super().__init__(path or '__x_ignore__')
        self._context = None
        if cde is None:
            if self.built():
                with open(self.path/'pt_meta.json') as fin:
                    model_name = json.load(fin)['cde_model']
                cde = CDE(model_name, cache=self)
            else:
                cde = CDE(cache=self)
        elif isinstance(cde, str):
            cde = CDE(cde, cache=self)
        self.cde = cde
        self.rng = rng

    def index(self, inp):
        assert not self.built()
        context_docs_size = self.cde.config.transductive_corpus_size
        sample = _sample_iter(inp, k=context_docs_size, rng=self.rng)
        self._context = self.cde.encode_context([d[self.cde.text_field] for d in sample])
        if str(self.path) != '__x_ignore__':
            with pta.ArtifactBuilder(self) as builder:
                builder.metadata['cde_model'] = self.cde.model_name
                builder.metadata['shape'] = list(self._context.shape)
                self._context.tofile(builder.path / 'sample.np')

    def __repr__(self):
        if str(self.path) == '__x_ignore__':
            return 'CDECache()'
        return f'CDECache({self.path!r})'

    def built(self):
        if str(self.path) == '__x_ignore__':
            return self._context is not None
        return os.path.exists(self.path)

    def context(self):
        if self._context is None:
            if not self.built():
                raise RuntimeError('{self!r} is not built. Run cde.cache.index() over your corpus first to build the context cache.')
            ctxt = np.fromfile(self.path/'sample.np', dtype=np.float32)
            with open(self.path/'pt_meta.json') as fin:
                shape = json.load(fin)['shape']
            self._context = ctxt.reshape(shape)
        return self._context



import os
import numpy as np
import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta
import pyterrier_dr
from pyterrier_dr import SimFn
from . import FlexIndex


class FlatNavRetriever(pt.Transformer):
    def __init__(self, flex_index, flatnav_index, *, threads=16, ef_search=100, num_initializations=100, num_results=1000, qbatch=64, drop_query_vec=False, verbose=False):
        self.flex_index = flex_index
        self.flatnav_index = flatnav_index
        self.threads = threads
        self.ef_search = ef_search
        self.num_results = num_results
        self.qbatch = qbatch
        self.drop_query_vec = drop_query_vec
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        docnos, config = self.flex_index.payload(return_dvecs=False)
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = query_vecs.copy()
        num_q = query_vecs.shape[0]
        QBATCH = self.qbatch
        it = range(0, num_q, QBATCH)
        if self.flex_index.verbose:
            it = pt.tqdm(it, unit='qbatch')
        self.flatnav_index.set_num_threads(self.threads)

        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        for qidx in it:
            scores, dids = self.flatnav_index.search(
                queries=query_vecs[qidx:qidx+QBATCH],
                ef_search=self.ef_search,
                K=min(self.num_results, len(self.flex_index)),
            )
            scores = -scores # distances -> scores
            for s, d in zip(scores, dids):
                mask = d != -1
                d = d[mask]
                s = s[mask]
                result.extend({
                    'docno': docnos.fwd[d],
                    'docid': d,
                    'score': s,
                    'rank': np.arange(d.shape[0]),
                })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


def _flatnav_retriever(self,
    k: int = 32,
    *,
    ef_search: int = 100,
    num_initializations: int = 100,
    ef_construction: int = 100,
    threads: int = 16,
    num_results: int = 1000,
    cache: bool = True,
    qbatch: int = 64,
    drop_query_vec: bool = False,
    verbose: bool = False,
) -> pt.Transformer:
    """Returns a retriever that searchers over a flatnav index.

    Args:
        k (int): the maximum number of edges per document in the index
        ef_search (int): the size of the list during searches. Higher values are slower but more accurate.
        num_initializations (int): the number of random initializations to use during search.
        ef_construction (int): the size of the list during graph construction. Higher values are slower but more accurate.
        threads (int): the number of threads to use
        num_results (int): the number of results to return per query
        cache (bool): whether to cache the index to disk
        qbatch (int): the number of queries to search at once
        drop_query_vec (bool): whether to drop the query_vec column after retrieval
        verbose (bool): whether to show progress bars

    .. versionadded:: 0.4.0

    .. note::
        This transformer requires the ``flatnav`` package to be installed. Instructions are available
        in the `flatnav repository <https://github.com/BlaiseMuhirwa/flatnav>`__.

    .. cite:: arxiv:2412.01940
        :citation: Munyampirwa et al. Down with the Hierarchy: The 'H' in HNSW Stands for "Hubs". arXiv 2024.
        :link: https://arxiv.org/abs/2412.01940
    """
    pyterrier_dr.util.assert_flatnav()
    import flatnav

    key = ('flatnav', k, ef_construction)
    index_name = f'{k}_ef-{ef_construction}-{str(self.sim_fn)}.flatnav'
    if key not in self._cache:
        dvecs, meta = self.payload(return_docnos=False)
        if not os.path.exists(self.index_path/index_name):
            distance_type = {
                SimFn.dot: 'angular',
            }[self.sim_fn]
            idx = flatnav.index.create(
                distance_type=distance_type,
                index_data_type=flatnav.data_type.DataType.float32,
                dim=dvecs.shape[1],
                dataset_size=dvecs.shape[0],
                max_edges_per_node=k,
                verbose=True,
                collect_stats=True,
            )
            idx.set_num_threads(threads)
            idx.add(data=np.array(dvecs), ef_construction=ef_construction)
            # for start_idx in pt.tqdm(range(0, dvecs.shape[0], 4096), desc='indexing flatnav', unit='batch'):
            #     idx.add(data=np.array(dvecs[start_idx:start_idx+4096]), ef_construction=ef_construction)
            if cache:
                idx.save(str(self.index_path/index_name))
            self._cache[key] = idx
        else:
            self._cache[key] = flatnav.index.IndexIPFloat.load_index(str(self.index_path/index_name))
            self._cache[key].set_data_type(flatnav.data_type.DataType.float32)
    return FlatNavRetriever(self, self._cache[key], threads=threads, ef_search=ef_search, num_initializations=num_initializations, num_results=num_results, qbatch=qbatch, drop_query_vec=drop_query_vec, verbose=verbose)
FlexIndex.flatnav_retriever = _flatnav_retriever



from pyterrier_dr.flex.core import FlexIndex, IndexingMode
from pyterrier_dr.flex import np_retr
from pyterrier_dr.flex import torch_retr
from pyterrier_dr.flex import corpus_graph
from pyterrier_dr.flex import faiss_retr
from pyterrier_dr.flex import flatnav_retr
from pyterrier_dr.flex import scann_retr
from pyterrier_dr.flex import ladr
from pyterrier_dr.flex import gar
from pyterrier_dr.flex import voyager_retr

__all__ = ["FlexIndex", "IndexingMode", "np_retr", "torch_retr", "corpus_graph", "faiss_retr", "flatnav_retr", "scann_retr", "ladr", "gar", "voyager_retr"]



from typing import Optional
import pyterrier as pt
import numpy as np
import pandas as pd
from .. import SimFn
from ..indexes import RankedLists
from . import FlexIndex
import pyterrier_alpha as pta


class NumpyRetriever(pt.Transformer):
    def __init__(self,
        flex_index: FlexIndex,
        *,
        num_results: int = 1000,
        batch_size: Optional[int] = None,
        drop_query_vec: bool = False
    ):
        self.flex_index = flex_index
        self.num_results = num_results
        self.batch_size = batch_size or 4096
        self.drop_query_vec = drop_query_vec

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        query_vecs = np.stack(inp['query_vec'])
        docnos, dvecs, config = self.flex_index.payload()
        if self.flex_index.sim_fn == SimFn.cos:
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
        elif self.flex_index.sim_fn == SimFn.dot:
            pass # nothing to do
        else:
            raise ValueError(f'{self.flex_index.sim_fn} not supported')
        num_q = query_vecs.shape[0]
        ranked_lists = RankedLists(self.num_results, num_q)
        batch_it = range(0, dvecs.shape[0], self.batch_size)
        if self.flex_index.verbose:
            batch_it = pt.tqdm(batch_it, desc='NumpyRetriever scoring', unit='docbatch')
        for idx_start in batch_it:
            doc_batch = dvecs[idx_start:idx_start+self.batch_size].T
            if self.flex_index.sim_fn == SimFn.cos:
                doc_batch = doc_batch / np.linalg.norm(doc_batch, axis=0, keepdims=True)
            scores = query_vecs @ doc_batch
            dids = np.arange(idx_start, idx_start+doc_batch.shape[1], dtype='i4').reshape(1, -1).repeat(num_q, axis=0)
            ranked_lists.update(scores, dids)

        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        for scores, dids in zip(*ranked_lists.results()):
            result.extend({
                'docno': docnos.fwd[dids],
                'docid': dids,
                'score': scores,
                'rank': np.arange(len(scores)),
            })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


class NumpyVectorLoader(pt.Transformer):
    def __init__(self, flex_index: FlexIndex):
        self.flex_index = flex_index

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        docids = self.flex_index._load_docids(inp)
        dvecs, config = self.flex_index.payload(return_docnos=False)
        return inp.assign(doc_vec=list(dvecs[docids]))


class NumpyScorer(pt.Transformer):
    def __init__(self, flex_index: FlexIndex, *, num_results: Optional[int] = None):
        self.flex_index = flex_index
        self.num_results = num_results

    def score(self, query_vecs, docids):
        dvecs, config = self.flex_index.payload(return_docnos=False)
        doc_vecs = dvecs[docids]
        if self.flex_index.sim_fn == SimFn.cos:
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
            doc_vecs = doc_vecs / np.linalg.norm(doc_vecs, axis=1, keepdims=True)
            return query_vecs @ doc_vecs.T
        elif self.flex_index.sim_fn == SimFn.dot:
            return query_vecs @ doc_vecs.T
        else:
            raise ValueError(f'{self.flex_index.sim_fn} not supported')

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        with pta.validate.any(inp) as v:
            v.columns(includes=['query_vec', 'docno'])
            v.columns(includes=['query_vec', 'docid'])
        inp = inp.reset_index(drop=True)

        res_idxs = []
        res_scores = []
        res_ranks = []
        for qid, df in pt.tqdm(inp.groupby('qid')):
            docids = self.flex_index._load_docids(df)
            query_vecs = df['query_vec'].iloc[0].reshape(1, -1)
            scores = self.score(query_vecs, docids).reshape(-1)
            index = df.index
            if self.num_results is not None and scores.shape[0] > self.num_results:
                idxs = np.argpartition(scores, -self.flex_index.num_results)[-self.num_results:]
                scores = scores[idxs]
                index = index[idxs]
            idxs = (-scores).argsort()
            res_scores.append(scores[idxs])
            res_idxs.append(index[idxs])
            res_ranks.append(np.arange(idxs.shape[0]))
        res_idxs = np.concatenate(res_idxs)
        res_scores = np.concatenate(res_scores)
        res_ranks = np.concatenate(res_ranks)
        res = pd.DataFrame({k: inp[k][res_idxs] for k in inp.columns if k not in ['score', 'rank']})
        res = res.assign(score=res_scores, rank=res_ranks)
        return res

def _np_vecs(self) -> np.ndarray:
    """Return the indexed vectors.

    Returns:
        :class:`numpy.ndarray`: The indexed vectors as a memory-mapped numpy array.
    """
    dvecs, meta = self.payload(return_docnos=False)
    return dvecs
FlexIndex.np_vecs = _np_vecs

def _np_retriever(self, *, num_results: int = 1000, batch_size: Optional[int] = None, drop_query_vec: bool = False):
    """Return a retriever that uses numpy to perform a brute force search over the index.

    The returned transformer expects a DataFrame with columns ``qid`` and ``query_vec``. It outpus
    a result frame containing the retrieved results.

    Args:
        num_results: The number of results to return per query.
        batch_size: The number of documents to score in each batch.
        drop_query_vec: Whether to drop the query vector from the output.

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses numpy to perform a brute force search.
    """
    return NumpyRetriever(self, num_results=num_results, batch_size=batch_size, drop_query_vec=drop_query_vec)
FlexIndex.np_retriever = _np_retriever
FlexIndex.retriever = _np_retriever # default retriever


def _np_vec_loader(self):
    """Return a transformer that loads indexed vectors.

    The returned transformer expects a DataFrame with columns ``docno``. It outputs a frame that
    includes a column ``doc_vec``, which contains the indexed vectors.

    Returns:
        :class:`~pyterrier.Transformer`: A transformer that loads indexed vectors.
    """
    return NumpyVectorLoader(self)
FlexIndex.np_vec_loader = _np_vec_loader
FlexIndex.vec_loader = _np_vec_loader # default vec_loader


def _np_scorer(self, *, num_results: Optional[int] = None) -> pt.Transformer:
    """Return a scorer that uses numpy to score (re-rank) results using indexed vectors.

    The returned transformer expects a DataFrame with columns ``qid``, ``query_vec`` and ``docno``.
    (If an internal ``docid`` column is provided, this will be used to speed up vector lookups.)

    This method uses memory-mapping to avoid loading the entire index into memory at once.

    Args:
        num_results: The number of results to return per query. If not provided, all resuls from the original fram are returned.

    Returns:
        :class:`~pyterrier.Transformer`: A transformer that scores query vectors with numpy.
    """
    return NumpyScorer(self, num_results=num_results)
FlexIndex.np_scorer = _np_scorer
FlexIndex.scorer = _np_scorer # default scorer



import pyterrier as pt
import pandas as pd
import numpy as np
import torch
import pyterrier_alpha as pta
from .biencoder import BiEncoder

class BGEM3(BiEncoder):
    def __init__(self, model_name='BAAI/bge-m3', batch_size=32, max_length=8192, text_field='text', verbose=False, device=None, use_fp16=False):
        super().__init__(batch_size=batch_size, text_field=text_field, verbose=verbose)
        self.model_name = model_name
        self.use_fp16 = use_fp16
        self.max_length = max_length
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        try:
            from FlagEmbedding import BGEM3FlagModel
        except ImportError:
            raise ImportError("BGE-M3 requires the FlagEmbedding package. You can install it using 'pip install pyterrier-dr[bgem3]'")
        
        self.model = BGEM3FlagModel(self.model_name, use_fp16=self.use_fp16, device=self.device)


    def __repr__(self):
        return f'BGEM3({repr(self.model_name)})'
    
    def encode_queries(self, texts, batch_size=None):
        return self.model.encode(list(texts), batch_size=batch_size, max_length=self.max_length,
                                  return_dense=True, return_sparse=False, return_colbert_vecs=False)['dense_vecs']

    def encode_docs(self, texts, batch_size=None):
        return self.model.encode(list(texts), batch_size=batch_size, max_length=self.max_length,
                                    return_dense=True, return_sparse=False, return_colbert_vecs=False)['dense_vecs']

    # Only does dense (single_vec) encoding
    def query_encoder(self, verbose=None, batch_size=None):
        return BGEM3QueryEncoder(self, verbose=verbose, batch_size=batch_size)
    def doc_encoder(self, verbose=None, batch_size=None):
        return BGEM3DocEncoder(self, verbose=verbose, batch_size=batch_size)
    
    # Does all three BGE-M3 encodings: dense, sparse and colbert(multivec)
    def query_multi_encoder(self, verbose=None, batch_size=None, return_dense=True, return_sparse=True, return_colbert_vecs=True):
        return BGEM3QueryEncoder(self, verbose=verbose, batch_size=batch_size, return_dense=return_dense, return_sparse=return_sparse, return_colbert_vecs=return_colbert_vecs)
    def doc_multi_encoder(self, verbose=None, batch_size=None, return_dense=True, return_sparse=True, return_colbert_vecs=True):
        return BGEM3DocEncoder(self, verbose=verbose, batch_size=batch_size, return_dense=return_dense, return_sparse=return_sparse, return_colbert_vecs=return_colbert_vecs)

class BGEM3QueryEncoder(pt.Transformer):
    def __init__(self, bge_factory: BGEM3, verbose=None, batch_size=None, max_length=None, return_dense=True, return_sparse=False, return_colbert_vecs=False):
        self.bge_factory = bge_factory
        self.verbose = verbose if verbose is not None else bge_factory.verbose
        self.batch_size = batch_size if batch_size is not None else bge_factory.batch_size
        self.max_length = max_length if max_length is not None else bge_factory.max_length

        self.dense = return_dense
        self.sparse = return_sparse
        self.multivecs = return_colbert_vecs
    
    def encode(self, texts):
        return self.bge_factory.model.encode(list(texts), batch_size=self.batch_size, max_length=self.max_length,
                             return_dense=self.dense, return_sparse=self.sparse, return_colbert_vecs=self.multivecs)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=['query'])

        # check if inp is empty
        if len(inp) == 0:
            if self.dense:
                inp = inp.assign(query_vec=[])
            if self.sparse:
                inp = inp.assign(query_toks=[])
            if self.multivecs:
                inp = inp.assign(query_embs=[])
            return inp

        it = inp['query'].values
        it, inv = np.unique(it, return_inverse=True)
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Queries', unit='query')
        bgem3_results = self.encode(it)

        if self.dense:
            inp = inp.assign(query_vec=[bgem3_results['dense_vecs'][i] for i in inv])
        if self.sparse:
            # for sparse convert ids to the actual tokens
            query_toks = self.bge_factory.model.convert_id_to_token(bgem3_results['lexical_weights'])
            inp = inp.assign(query_toks=query_toks)
        if self.multivecs:
            inp = inp.assign(query_embs=[bgem3_results['colbert_vecs'][i] for i in inv])
        return inp
    
    def __repr__(self):
        return f'{repr(self.bge_factory)}.query_encoder()'

class BGEM3DocEncoder(pt.Transformer):
    def __init__(self, bge_factory: BGEM3, verbose=None, batch_size=None, max_length=None, return_dense=True, return_sparse=False, return_colbert_vecs=False):
        self.bge_factory = bge_factory
        self.verbose = verbose if verbose is not None else bge_factory.verbose
        self.batch_size = batch_size if batch_size is not None else bge_factory.batch_size
        self.max_length = max_length if max_length is not None else bge_factory.max_length

        self.dense = return_dense
        self.sparse = return_sparse
        self.multivecs = return_colbert_vecs

    def encode(self, texts):
        return self.bge_factory.model.encode(list(texts), batch_size=self.batch_size, max_length=self.max_length,
                             return_dense=self.dense, return_sparse=self.sparse, return_colbert_vecs=self.multivecs)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        # check if the input dataframe contains the field(s) specified in the text_field
        pta.validate.columns(inp, includes=[self.bge_factory.text_field])
        # check if inp is empty
        if len(inp) == 0:
            if self.dense:
                inp = inp.assign(doc_vec=[])
            if self.sparse:
                inp = inp.assign(toks=[])
            if self.multivecs:
                inp = inp.assign(doc_embs=[])
            return inp

        it = inp[self.bge_factory.text_field]
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Documents', unit='doc')
        bgem3_results = self.encode(it)

        if self.dense:
            inp = inp.assign(doc_vec=list(bgem3_results['dense_vecs']))
        if self.sparse:
            toks = bgem3_results['lexical_weights']
            # for sparse convert ids to the actual tokens
            toks = self.bge_factory.model.convert_id_to_token(toks)
            inp = inp.assign(toks=toks)
        if self.multivecs:
            inp = inp.assign(doc_embs=list(bgem3_results['colbert_vecs']))
        return inp

    def __repr__(self):
        return f'{repr(self.bge_factory)}.doc_encoder()'



import os
import pyterrier as pt
import numpy as np
import ir_datasets
import pyterrier_alpha as pta
import pyterrier_dr
from . import FlexIndex

logger = ir_datasets.log.easy()


class VoyagerRetriever(pt.Indexer):
    def __init__(self, flex_index, voyager_index, query_ef=None, num_results=1000, qbatch=64, drop_query_vec=False):
        self.flex_index = flex_index
        self.voyager_index = voyager_index
        self.query_ef = query_ef
        self.num_results = num_results
        self.qbatch = qbatch
        self.drop_query_vec = drop_query_vec

    def transform(self, inp):
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        docnos, config = self.flex_index.payload(return_dvecs=False)
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = query_vecs.copy()
        
        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        num_q = query_vecs.shape[0]
        QBATCH = self.qbatch
        it = range(0, num_q, QBATCH)
        if self.flex_index.verbose:
            it = pt.tqdm(it, unit='qbatch')
        for qidx in it:
            qvec_batch = query_vecs[qidx:qidx+QBATCH]
            neighbor_ids, distances = self.voyager_index.query(qvec_batch, self.num_results, self.query_ef)
            for s, d in zip(distances, neighbor_ids):
                mask = d != -1
                d = d[mask]
                s = s[mask]
                result.extend({
                    'docno': docnos.fwd[d],
                    'docid': d,
                    'score': -s,
                    'rank': np.arange(d.shape[0]),
                })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


def _voyager_retriever(self,
    neighbours: int = 12,
    *,
    num_results: int = 1000,
    ef_construction: int = 200,
    random_seed: int = 1,
    storage_data_type: str = 'float32',
    query_ef: int = 10,
    drop_query_vec: bool = False
) -> pt.Transformer:
    """Returns a retriever that uses HNSW to search over a Voyager index.

    Args:
        neighbours (int, optional): Number of neighbours to search. Defaults to 12.
        num_results (int, optional): Number of results to return per query. Defaults to 1000.
        ef_construction (int, optional): Expansion factor for graph construction. Defaults to 200.
        random_seed (int, optional): Random seed. Defaults to 1.
        storage_data_type (str, optional): Storage data type. One of 'float32', 'float8', 'e4m3'. Defaults to 'float32'.
        query_ef (int, optional): Expansion factor during querying. Defaults to 10.
        drop_query_vec (bool, optional): Drop the query vector from the output. Defaults to False.

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses HNSW to search over a Voyager index.

    .. note::
        This method requires the ``voyager`` package. Install it via ``pip install voyager``.
    """
    pyterrier_dr.util.assert_voyager()
    import voyager
    meta, = self.payload(return_dvecs=False, return_docnos=False)

    space = {
        pyterrier_dr.SimFn.dot: voyager.Space.InnerProduct,
        pyterrier_dr.SimFn.cos: voyager.Space.Cosine,
    }[self.sim_fn]
    storage_data_type = {
        'float32': voyager.StorageDataType.Float32,
        'float8': voyager.StorageDataType.Float8,
        'e4m3': voyager.StorageDataType.E4M3,
    }[storage_data_type.lower()]

    key = ('voyager', space, neighbours, ef_construction, random_seed, storage_data_type)
    index_name = f'voyager-{space.name}-{neighbours}-{ef_construction}-{random_seed}-{storage_data_type.name}.voyager'
    if key not in self._cache:
        if not os.path.exists(self.index_path/index_name):
            BATCH_SIZE = 10_000
            dvecs, meta = self.payload(return_docnos=False)
            index = voyager.Index(space, meta['vec_size'], neighbours, ef_construction, random_seed, meta['doc_count'], storage_data_type)
            print(index.ef)
            it = range(0, meta['doc_count'], BATCH_SIZE)
            if self.verbose:
                it = pt.tqdm(it, desc='building index', unit='dbatch')
            for idx in it:
                index.add_items(dvecs[idx:idx+BATCH_SIZE])
            with logger.duration('saving index'):
                index.save(str(self.index_path/index_name))
            self._cache[key] = index
        else:
            with logger.duration('reading index'):
                self._cache[key] = voyager.Index.load(str(self.index_path/index_name))
    return VoyagerRetriever(self, self._cache[key], query_ef=query_ef, drop_query_vec=drop_query_vec)
FlexIndex.voyager_retriever = _voyager_retriever



import re
import math
import json
import ir_datasets
import torch
import numpy as np
import pyterrier as pt
import pyterrier_dr
from ..indexes import TorchRankedLists
from . import FlexIndex


def _corpus_graph(self, k: int = 16, *, batch_size: int = 8192):
    """Return the corpus graph (neighborhood graph) for the index.

    The corpus graph is a directed graph where each node represents a document and each edge represents a
    connection between two documents. The graph is built by computing the cosine similarity between each
    pair of documents and storing the k-nearest neighbors for each document.

    If the corpus graph has not been built yet, it will be built using the given k and batch size.

    Args:
        k: The number of neighbors to store for each document.
        batch_size: The number of vectors to process in each batch.

    Returns:
        :class:`pyterrier_adaptive.CorpusGraph`: The corpus graph for the index.
    """
    from pyterrier_adaptive import CorpusGraph
    key = ('corpus_graph', k)
    if key not in self._cache:
        path = self.index_path/f'corpusgraph_k{k}'
        if not (path/'pt_meta.json').exists():
            candidates = [(p, re.match(r'^corpusgraph_k([0-9]+)$', str(p).split('/')[-2])) for p in self.index_path.glob('corpusgraph_k*/pt_meta.json')]
            candidates = sorted([(int(m.group(1)), p) for p, m in candidates if m is not None and int(m.group(1)) > k])
            if candidates:
                self._cache[key] = CorpusGraph.load(candidates[0][1].parent).to_limit_k(k)
            else:
                _build_corpus_graph(self, k, path, batch_size)
                self._cache[key] = CorpusGraph.load(path)
        else:
            self._cache[key] = CorpusGraph.load(path)
    return self._cache[key]


FlexIndex.corpus_graph = _corpus_graph


def _build_corpus_graph(flex_index, k, out_dir, batch_size):
    S = batch_size
    vectors, meta = flex_index.payload(return_docnos=False)
    num_vecs = vectors.shape[0]
    num_chunks = math.ceil(num_vecs/S)
    rankings = [TorchRankedLists(k, min((i+1)*S, num_vecs) - i*S) for i in range(num_chunks)]
    out_dir.mkdir(parents=True, exist_ok=True)
    edges_path = out_dir/'edges.u32.np'
    weights_path = out_dir/'weights.f16.np'
    device = pyterrier_dr.util.infer_device()
    dtype = torch.half if device.type == 'cuda' else torch.float
    with pt.tqdm(total=int((num_chunks+1)*num_chunks/2), unit='chunk', smoothing=1) as pbar, \
        ir_datasets.util.finialized_file(str(edges_path), 'wb') as fe, \
        ir_datasets.util.finialized_file(str(weights_path), 'wb') as fw:
        for i in range(num_chunks):
            left = torch.from_numpy(vectors[i*S:(i+1)*S]).to(device).to(dtype)
            left = left / left.norm(dim=1, keepdim=True)
            scores = left @ left.T
            scores[torch.eye(left.shape[0], dtype=bool, device=device)] = float('-inf')
            i_scores, i_dids = scores.topk(k, sorted=True, dim=1)
            rankings[i].update(i_scores, (i_dids + i*S))
            pbar.update()
            for j in range(i+1, num_chunks):
                right = torch.from_numpy(vectors[j*S:(j+1)*S]).to(device).to(dtype)
                right = right / right.norm(dim=1, keepdim=True)
                scores = left @ right.T
                i_scores, i_dids = scores.topk(min(k, right.shape[0]), sorted=True, dim=1)
                j_scores, j_dids = scores.topk(min(k, left.shape[0]), sorted=True, dim=0)
                rankings[i].update(i_scores, (i_dids + j*S))
                rankings[j].update(j_scores.T, (j_dids + i*S).T)
                pbar.update()
            sims, idxs = rankings[i].results()
            fe.write(idxs.astype(np.uint32).tobytes())
            fw.write(sims.astype(np.float16).tobytes())
            rankings[i] = None
    (out_dir/'docnos.npids').symlink_to('../docnos.npids')
    with (out_dir/'pt_meta.json').open('wt') as fout:
      json.dump({
        'type': 'corpus_graph',
        'format': 'np_topk',
        'doc_count': vectors.shape[0],
        'k': k,
      }, fout)



import numpy as np
import more_itertools
import torch
import pyterrier as pt
from transformers import AutoTokenizer, ElectraForSequenceClassification

class ElectraScorer(pt.Transformer):
    def __init__(self, model_name='crystina-z/monoELECTRA_LCE_nneg31', batch_size=16, text_field='text', verbose=True, device=None):
        self.model_name = model_name
        self.batch_size = batch_size
        self.text_field = text_field
        self.verbose = verbose
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.tokeniser = AutoTokenizer.from_pretrained('google/electra-base-discriminator')
        self.model = ElectraForSequenceClassification.from_pretrained(model_name).eval().to(self.device)

    def transform(self, inp):
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='ELECTRA scoring')
        with torch.no_grad():
            for chunk in more_itertools.chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                inps = self.tokeniser(queries, texts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                scores.append(self.model(**inps).logits[:, 1].cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        pt.model.add_ranks(res)
        res = res.sort_values(['qid', 'rank'])
        return res



# Deprecated module
import torch
import itertools
import math
import json
from pathlib import Path
from os.path import commonprefix
import numpy as np
import shutil
import more_itertools
import pandas as pd
import pyterrier as pt
from pyterrier.model import add_ranks
import ir_datasets

_logger = ir_datasets.log.easy()


class DocnoFile:
    def __init__(self, path):
        with open(path, 'rb') as f:
            metadata_size = int.from_bytes(f.read(2), byteorder='little')
            metadata = json.loads(f.read(metadata_size))
        self.prefix = metadata['prefix']
        self.max_length = metadata['max_length']
        self.count = metadata['count']
        if 'start' in metadata:
            self.start = metadata['start']
            self.step = metadata['step']
            self.fmt_string = metadata['fmt_string']
            self.mmap = None
        else:
            self.mmap = np.memmap(path, f'S{self.max_length}', offset=metadata_size+2)

    def __getitem__(self, idx):
        if isinstance(idx, (int, np.int64, np.int32)):
            if self.mmap is not None:
                return self.prefix + self.mmap[idx].decode()
            else:
                assert 0 <= idx < self.count
                return self._fmt_idx(idx)
        elif isinstance(idx, str):
            if self.mmap is not None:
                raise RuntimeError("reverse lookup not supported")
            else:
                if not idx.startswith(self.prefix):
                    return None
                idx = idx[len(self.prefix):]
                if not idx.isnumeric():
                    return None
                idx = int(idx)
                if idx % self.step != 0:
                    return None
                idx = (idx // self.step) - self.start
                if idx < 0 or idx >= self.count:
                    return None
                return idx
        else:
            if self.mmap is not None:
                docnos = self.mmap[idx]
                return np.array([self.prefix + d.decode() for d in docnos], dtype=object)
            else:
                assert ((0 <= idx) & (idx < self.count)).all()
                docnos = idx * self.step + self.start
                return np.array([self.prefix + format(d, self.fmt_string) for d in docnos], dtype=object)

    def __len__(self):
        return self.count

    def __iter__(self):
        if self.mmap is not None:
            for item in self.mmap:
                yield self.prefix + item.decode()
        else:
            for idx in range(self.count):
                yield self._fmt_idx(idx)

    def __del__(self):
        del self.mmap
        self.mmap = None

    def _fmt_idx(self, idx):
        return self.prefix + format(idx * self.step + self.start, self.fmt_string)

    @staticmethod
    def build(docnos, path):
        assert len(docnos) > 0
        prefix = commonprefix(docnos) # reduce the size by removing common docno prefixes
        docnos = [d[len(prefix):].encode() for d in docnos] # remove prefix
        max_length = max(len(d) for d in docnos)
        metadata = {'count': len(docnos), 'prefix': prefix, 'max_length': max_length}
        if all(d.decode().isnumeric() for d in docnos) and len(docnos) >= 2:
            # Special case: are the docnos (1) all numeric, (2) incrementing at a constant step?
            start = int(docnos[0])
            step = int(docnos[1]) - start
            fmt_string = f'0{max_length}d' if any(d!=b'0' and d.startswith(b'0') for d in docnos) else 'd'
            for d, expected in zip(docnos, itertools.count(start, step)):
                if d.decode() != format(expected, fmt_string):
                    break # pattern broken
            else:
                # everything was numeric and incremental
                metadata['start'] = start
                metadata['step'] = step
                metadata['fmt_string'] = fmt_string
        with open(path, 'wb') as f:
            b_metadata = json.dumps(metadata).encode()
            f.write(len(b_metadata).to_bytes(2, byteorder='little'))
            f.write(b_metadata)
            if 'start' not in metadata:
                f.write(np.array(docnos, dtype=f'S{max_length}').tobytes())


class NilIndex(pt.Indexer):
    # simulates an indexer; just used for testing
    def transform(self, res):
        return res

    def index(self, it):
        for _ in it:
            pass


class NumpyIndex(pt.Indexer):
    def __init__(self, index_path=None, num_results=1000, score_fn='dot', overwrite=False, batch_size=4096, verbose=False, dtype='f4', drop_query_vec=True, inmem=False, cuda=False):
        self.index_path = Path(index_path)
        self.num_results = num_results
        self.score_fn = score_fn
        self.overwrite = overwrite
        self.verbose = verbose
        self.batch_size = batch_size
        self.dtype = dtype
        self.drop_query_vec = drop_query_vec
        self._docnos = None
        self._data = None
        self.inmem = inmem
        self.cuda = cuda

    def docnos_and_data(self):
        if self._data is None:
            path = self.index_path
            with open(path/'meta.json', 'rt') as f_meta:
                meta = json.load(f_meta)
            if self.inmem:
                with open(path/'index.npy', 'rb') as fin:
                    self._data = np.frombuffer(fin.read(), dtype=self.dtype).reshape(meta['count'], meta['vec_size'])
                if self.cuda:
                    self._data = torch.tensor(self._data, dtype=torch.float16, device='cuda')
            else:
                self._data = np.memmap(path/'index.npy', mode='r', dtype=self.dtype, shape=(meta['count'], meta['vec_size']))
            self._docnos = DocnoFile(path/'docnos.npy')
        return self._docnos, self._data

    def __len__(self):
        with open(self.index_path/'meta.json', 'rt') as f_meta:
            meta = json.load(f_meta)
        return meta['count']

    def transform(self, inp):
        columns = set(inp.columns)
        if all(c in columns for c in ('qid', 'query_vec', 'docno')):
            return self.transform_R(inp)
        if all(c in columns for c in ('qid', 'query_vec')):
            return self.transform_Q(inp)
        raise RuntimeError('NumpyIndex expects input columns of either:\n'
            ' - ["qid", "query_vec", "docno", ...]: Re-ranking\n'
            ' - ["qid", "query_vec", ...]: Retrieval\n'
            '(Are you encoding the query?)')

    def transform_Q(self, inp):
        query_vecs = np.stack(inp['query_vec'])
        if self.score_fn == 'cos':
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
        if self.cuda:
            query_vecs = torch.from_numpy(query_vecs).cuda()
        num_q = query_vecs.shape[0]
        res = []
        docnos, data = self.docnos_and_data()
        ranked_lists = RankedLists(self.num_results, num_q)
        if self.cuda and self.inmem:
            scores = query_vecs.half() @ data.T
            scores, dids = torch.topk(scores, k=self.num_results, dim=1)
            scores = scores.cpu().float().numpy()
            dids = dids.cpu().numpy()
            ranked_lists.update(scores, dids)
        else:
            batch_it = range(0, data.shape[0], self.batch_size)
            if self.verbose:
                batch_it = pt.tqdm(batch_it)
            for idx_start in batch_it:
                doc_batch = data[idx_start:idx_start+self.batch_size].T
                if self.score_fn == 'cos':
                    doc_batch = doc_batch / np.linalg.norm(doc_batch, axis=0, keepdims=True)
                if self.cuda:
                    doc_batch = torch.from_numpy(doc_batch).cuda()
                scores = query_vecs @ doc_batch
                if self.cuda:
                    scores = scores.cpu().numpy()
                dids = np.arange(idx_start, idx_start+doc_batch.shape[1], dtype='i4').reshape(1, -1).repeat(num_q, axis=0)
                ranked_lists.update(scores, dids)
        result_scores, result_dids = ranked_lists.results()
        unique_dids, unique_inverse = np.unique(result_dids, return_inverse=True)
        unique_docnos = docnos[unique_dids]
        result_docnos = unique_docnos[unique_inverse].reshape(num_q, -1)
        for query, scores, docnos in zip(inp.itertuples(index=False), result_scores, result_docnos):
            if self.drop_query_vec:
                query = query._replace(query_vec=None)
            for score, docno in zip(scores, docnos):
                res.append((*query, docno, score))
        res = pd.DataFrame(res, columns=list(query._fields) + ['docno', 'score'])
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = res[~res.score.isna()]
        res = add_ranks(res)
        return res

    def transform_R(self, inp):
        assert not self.cuda, "cuda not yet supported for re-ranking"
        docnos, data = self.docnos_and_data()
        all_scores = []
        it = range(0, len(inp), self.batch_size)
        if self.verbose:
            it = pt.tqdm(it, desc='re-ranking from index', unit='batch')
        for start_idx in it:
            end_idx = start_idx + self.batch_size
            query_vecs = np.stack(inp.iloc[start_idx:end_idx]['query_vec'])
            dids = np.array([docnos[str(d)] for d in inp.iloc[start_idx:end_idx]['docno']]) # reverse lookup
            doc_vecs = data[dids]
            if self.score_fn == 'cos':
                query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
                doc_vecs = doc_vecs / np.linalg.norm(doc_vecs, axis=1, keepdims=True)
            all_scores.append((query_vecs * doc_vecs).sum(axis=1))
        all_scores = np.concatenate(all_scores, axis=0)
        res = inp.assign(score=all_scores)
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = res[~res.score.isna()]
        res = add_ranks(res)
        return res

    def index(self, inp):
        if isinstance(inp, pd.DataFrame):
            inp = inp.to_dict(orient="records")
        inp = more_itertools.peekable(inp)
        path = Path(self.index_path)
        if path.exists():
            if self.overwrite:
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Index already exists at {self.index_path}. If you want to delete and re-create an existing index, you can pass overwrite=True')
        path.mkdir(parents=True, exist_ok=True)
        docnos = []
        vec_size = None
        count = 0
        with open(path/'index.npy', 'wb') as fout:
            for docs in more_itertools.chunked(inp, self.batch_size):
                doc_vecs = np.stack([d['doc_vec'] for d in docs])
                if vec_size is None:
                    vec_size = doc_vecs.shape[1]
                elif vec_size != doc_vecs.shape[1]:
                    raise ValueError('Inconsistent vector shapes detected')
                doc_vecs = doc_vecs.astype(self.dtype)
                fout.write(doc_vecs.tobytes())
                docnos.extend([d['docno'] for d in docs])
                count += len(docs)
        DocnoFile.build(docnos, path/'docnos.npy')
        with open(path/'meta.json', 'wt') as f_meta:
            json.dump({'dtype': self.dtype, 'vec_size': vec_size, 'count': count}, f_meta)

    def get_corpus_iter(self, start_idx=None, stop_idx=None, verbose=True):
        docnos, data = self.docnos_and_data()
        docno_iter = iter(docnos)
        if start_idx is not None or stop_idx is not None:
            docno_iter = itertools.islice(docno_iter, start_idx, stop_idx)
            data = data[start_idx:stop_idx]
        with pt.tqdm(total=data.shape[0], desc=f'{str(self.index_path)} document vectors') as pbar:
            for start_idx in range(0, data.shape[0], self.batch_size):
                vec_batch = data[start_idx:start_idx+self.batch_size]
                for vec_idx in range(vec_batch.shape[0]):
                    pbar.update()
                    yield {'docno': next(docno_iter), 'doc_vec': vec_batch[vec_idx]}


class MemIndex(pt.Indexer):
    def __init__(self, num_results=1000, score_fn='dot', batch_size=4096, verbose=True, dtype='f4', drop_query_vec=True):
        # check we havent been passed a destination index, as per disk-based indexers
        assert isinstance(num_results, int)
        self.num_results = num_results
        self.score_fn = score_fn
        self.verbose = verbose
        self.batch_size = batch_size
        self.dtype = dtype
        self.drop_query_vec = drop_query_vec
        self._docnos = None
        self._data = None

    def docnos_and_data(self):
        return self._docnos, self._data

    def clear(self):
        self._docnos = None
        self._data = None

    def __len__(self):
        return self._data.shape[0]

    def transform(self, inp):
        columns = set(inp.columns)
        assert all(f in columns for f in ['qid', 'query_vec']), "NumpyIndex expects columns: ['qid', 'query_vec'] when used in a pipeline"
        query_vecs = np.stack(inp['query_vec'])
        if self.score_fn == 'cos':
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
        num_q = query_vecs.shape[0]
        res = []
        docnos, data = self.docnos_and_data()
        data = data.T
        assert docnos is not None or data is not None, "You must first call index()"
        ranked_lists = RankedLists(self.num_results, num_q)
        if self.score_fn == 'cos':
            data = data / np.linalg.norm(data, axis=0, keepdims=True)
        scores = query_vecs @ data
        dids = np.arange(data.shape[1], dtype='i4').reshape(1, -1).repeat(num_q, axis=0)
        ranked_lists.update(scores, dids)
        result_scores, result_dids = ranked_lists.results()
        for query, scores, dids in zip(inp.itertuples(index=False), result_scores, result_dids):
            if self.drop_query_vec:
                query = query._replace(query_vec=None)
            for score, did in zip(scores, dids):
                res.append((*query, docnos[int(did)], score))
        res = pd.DataFrame(res, columns=list(query._fields) + ['docno', 'score'])
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = res[~res.score.isna()]
        res = add_ranks(res)
        return res

    def index(self, inp):
        if isinstance(inp, pd.DataFrame):
            inp = inp.to_dict(orient="records")
        inp = more_itertools.peekable(inp)
        if self.verbose:
            inp = pt.tqdm(inp, desc='indexing', unit='doc')
        docnos = []
        vec_size = None
        count = 0
        data = []
        for docs in more_itertools.chunked(inp, self.batch_size):
            doc_vecs = np.stack([d['doc_vec'] for d in docs])
            if vec_size is None:
                vec_size = doc_vecs.shape[1]
            doc_vecs = doc_vecs.astype(self.dtype)
            data.append(doc_vecs)
            docnos.extend([d['docno'] for d in docs])
            count += len(docs)
        self._data = np.concatenate(data, axis=0)
        self._docnos = docnos

    def get_corpus_iter(self, start_idx=None, stop_idx=None, verbose=True):
        docnos, data = self.docnos_and_data()
        docno_iter = iter(docnos)
        if start_idx is not None or stop_idx is not None:
            docno_iter = iter(docnos[start_idx:stop_idx])
            data = data[start_idx:stop_idx]
        with pt.tqdm(total=data.shape[0], desc=f'{str(self.index_path)} document vectors') as pbar:
            for start_idx in range(0, data.shape[0], self.batch_size):
                vec_batch = data[start_idx:start_idx+self.batch_size]
                for vec_idx in range(vec_batch.shape[0]):
                    pbar.update()
                    yield {'docno': next(docno_iter), 'doc_vec': vec_batch[vec_idx]}


class RankedLists:
    def __init__(self, num_results : int, num_queries : int):
        self.num_results = num_results
        self.num_queries = num_queries
        self.scores = np.empty((num_queries, 0), dtype='f4')
        self.docids = np.empty((num_queries, 0), dtype='i4')

    def update(self, scores, docids):
        assert self.num_queries == scores.shape[0]
        self.scores = np.concatenate([self.scores, -scores], axis=1)
        self.docids = np.concatenate([self.docids,  docids], axis=1)
        if self.scores.shape[1] > self.num_results:
            partition_idxs = np.argpartition(self.scores, self.num_results, axis=1)[:, :self.num_results]
            self.scores = np.take_along_axis(self.scores, partition_idxs, axis=1)
            self.docids = np.take_along_axis(self.docids, partition_idxs, axis=1)

    def results(self):
        sort_idxs = np.argsort(self.scores, axis=1)
        return (
            np.take_along_axis(-self.scores, sort_idxs, axis=1),
            np.take_along_axis( self.docids, sort_idxs, axis=1),
        )


class TorchRankedLists:
    def __init__(self, num_results, num_queries):
        self.num_results = num_results
        self.num_queries = num_queries
        self.scores = None
        self.docids = None

    def update(self, scores, docids):
        if self.scores is not None:
            self.scores, idxs = torch.cat([self.scores, scores], dim=1).topk(self.num_results, dim=1)
            self.docids = torch.cat([self.docids, docids], dim=1).take_along_dim(idxs, dim=1)
        else:
            self.scores = scores
            self.docids = docids
        #self.scores.append(scores)
        #self.docids.append(docids)

    def results(self):
        return (self.scores.cpu().numpy(), self.docids.cpu().numpy())
        #scores = torch.cat(self.scores, dim=0)
        #docids = torch.cat(self.docids, dim=0)
        #top_scores, top_idxs = scores.topk(self.num_results, dim=1)
        #return (
        #    top_scores.cpu().numpy(),
        #    docids[top_idxs].cpu().numpy()
        #)


class FaissFlat(pt.Indexer):
    def __init__(self, index_path=None, num_results=1000, shard_size=500_000, score_fn='cos', overwrite=False, batch_size=4096, verbose=False, drop_query_vec=True, inmem=False, cuda=False):
        self.index_path = Path(index_path)
        self.num_results = num_results
        self.shard_size = shard_size
        self.score_fn = score_fn
        self.overwrite = overwrite
        self.verbose = verbose
        self.batch_size = batch_size
        self.drop_query_vec = drop_query_vec
        self.cuda = cuda
        self._shards = None
        self._shards = list(self.iter_shards())

    def iter_shards(self):
        import faiss
        if self.cuda:
            res = faiss.StandardGpuResources()
        if self._shards is None:
            for shardid in itertools.count():
                if not (self.index_path/f'{shardid}.faiss').exists():
                    break
                index = faiss.read_index(str(self.index_path/f'{shardid}.faiss'))
                if self.cuda:
                    index = faiss.index_cpu_to_gpu(res, 0, index)
                yield index
        else:
            yield from self._shards

        # if self._shards is None:
        #     shards = []
        #     for shardid in itertools.count():
        #         if not (self.index_path/f'{shardid}.faiss').exists():
        #             break
        #         index = faiss.read_index(str(self.index_path/f'{shardid}.faiss'))
        #         shards.append(index)
        #     self._shards = shards
        #     if shardid == 0:
        #         raise RuntimeError(f'Index not found at {self.index_path}')
        # return iter(self._shards)

    def transform(self, inp):
        columns = set(inp.columns)
        assert all(f in columns for f in ['qid', 'query_vec']), "Faiss expects columns: ['qid', 'query_vec'] when used in a pipeline"
        query_vecs = np.stack(inp['query_vec'])
        if self.score_fn == 'cos':
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
        query_vecs = query_vecs.copy()
        res = []
        docnos = DocnoFile(self.index_path/'docnos.npy')
        num_q = query_vecs.shape[0]
        ranked_lists = RankedLists(self.num_results, num_q)
        dids_offset = 0
        it = enumerate(self.iter_shards())
        if self.verbose:
            it = pt.tqdm(it, unit='shard', desc='scanning shards')
        for shardidx, index in it:
            scores, dids = [], []
            QBATCH = 16
            for qidx in range(0, num_q, QBATCH):
                s, d = index.search(query_vecs[qidx:qidx+QBATCH], self.num_results)
                scores.append(s)
                dids.append(d)
            ranked_lists.update(np.concatenate(scores, axis=0), np.concatenate(dids, axis=0)+dids_offset)
            dids_offset += index.ntotal
        result_scores, result_dids = ranked_lists.results()
        unique_dids, unique_inverse = np.unique(result_dids, return_inverse=True)
        unique_docnos = docnos[unique_dids]
        result_docnos = unique_docnos[unique_inverse].reshape(num_q, -1)
        for query, scores, docnos in zip(inp.itertuples(), result_scores, result_docnos):
            if self.drop_query_vec:
                query = query._replace(query_vec=None)
            for score, docno in zip(scores, docnos):
                res.append((*query, docno, score))
        res = pd.DataFrame(res, columns=list(query._fields) + ['docno', 'score'])
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = add_ranks(res)
        return res

    def index(self, inp):
        import faiss
        if isinstance(inp, pd.DataFrame):
            inp = inp.to_dict(orient="records")
        path = Path(self.index_path)
        if path.exists():
            if self.overwrite:
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Index already exists at {self.index_path}. If you want to delete and re-create an existing index, you can pass overwrite=True')
        path.mkdir(parents=True, exist_ok=True)
        docnos = []
        for shardid, shard in enumerate(more_itertools.ichunked(inp, self.shard_size)):
            shard = more_itertools.peekable(shard)
            d = shard.peek()['doc_vec'].shape[0]
            index = faiss.index_factory(d, 'Flat', faiss.METRIC_INNER_PRODUCT)
            shard_it = more_itertools.chunked(shard, self.batch_size)
            if self.verbose:
                shard_it = pt.tqdm(shard_it, total=math.ceil(self.shard_size/self.batch_size), unit='batch', desc=f'building shard {shardid}')
            for batch in shard_it:
                doc_vecs = np.stack(d['doc_vec'] for d in batch)
                if self.score_fn == 'cos':
                    doc_vecs = doc_vecs / np.linalg.norm(doc_vecs, axis=1, keepdims=True)
                index.add(doc_vecs)
                docnos.extend(d['docno'] for d in batch)
            faiss.write_index(index, str(path/f'{shardid}.faiss'))
        DocnoFile.build(docnos, path/'docnos.npy')


class FaissHnsw(pt.Indexer):
    def __init__(self, index_path, num_links=32, num_results=1000, shard_size=500_000, score_fn='cos', overwrite=False, batch_size=4096, verbose=False, drop_query_vec=True, qbatch=16):
        self.index_path = Path(index_path)
        self.num_links = num_links
        self.num_results = num_results
        self.shard_size = shard_size
        self.score_fn = score_fn
        self.overwrite = overwrite
        self.verbose = verbose
        self.batch_size = batch_size
        self.drop_query_vec = drop_query_vec
        self.qbatch = qbatch
        self._shards = None

    def iter_shards(self, cache=True):
        if not cache:
            return self._iter_shards()
        else:
            if self._shards is None:
                self._shards = list(self._iter_shards())
            return iter(self._shards)

    def _iter_shards(self):
        import faiss
        for shardid in itertools.count():
            if not (self.index_path/f'{shardid}.faiss').exists():
                break
            yield faiss.read_index(str(self.index_path/f'{shardid}.faiss'))
        if shardid == 0:
            raise RuntimeError(f'Index not found at {self.index_path}')

    def transform(self, inp):
        columns = set(inp.columns)
        assert all(f in columns for f in ['qid', 'query_vec']), "Faiss expects columns: ['qid', 'query_vec'] when used in a pipeline"
        query_vecs = np.stack(inp['query_vec'])
        if self.score_fn == 'cos':
            query_vecs = query_vecs / np.linalg.norm(query_vecs, axis=1, keepdims=True)
        query_vecs = query_vecs.copy()
        res = []
        docnos = DocnoFile(self.index_path/'docnos.npy')
        num_q = query_vecs.shape[0]
        ranked_lists = RankedLists(self.num_results, num_q)
        dids_offset = 0
        it = enumerate(self.iter_shards())
        if self.verbose:
            it = pt.tqdm(it, unit='shard', desc='scanning shards')
        for shardidx, index in it:
            scores, dids = [], []
            for qidx in range(0, num_q, self.qbatch):
                s, d = index.search(query_vecs[qidx:qidx+self.qbatch], self.num_results)
                scores.append(s)
                dids.append(d)
            ranked_lists.update(np.concatenate(scores, axis=0), np.concatenate(dids, axis=0)+dids_offset)
            dids_offset += index.ntotal
        result_scores, result_dids = ranked_lists.results()
        unique_dids, unique_inverse = np.unique(result_dids, return_inverse=True)
        unique_docnos = docnos[unique_dids]
        result_docnos = unique_docnos[unique_inverse].reshape(num_q, -1)
        for query, scores, docnos in zip(inp.itertuples(), result_scores, result_docnos):
            if self.drop_query_vec:
                query = query._replace(query_vec=None)
            for score, docno in zip(scores, docnos):
                res.append((*query, docno, score))
        res = pd.DataFrame(res, columns=list(query._fields) + ['docno', 'score'])
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = add_ranks(res)
        return res

    def index(self, inp):
        import faiss
        if isinstance(inp, pd.DataFrame):
            inp = inp.to_dict(orient="records")
        path = Path(self.index_path)
        if path.exists():
            if self.overwrite:
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Index already exists at {self.index_path}. If you want to delete and re-create an existing index, you can pass overwrite=True')
        path.mkdir(parents=True, exist_ok=True)
        docnos = []
        if self.shard_size == 0:
            it = [(0, inp)]
        else:
            it = enumerate(more_itertools.ichunked(inp, self.shard_size))
        for shardid, shard in it:
            shard = more_itertools.peekable(shard)
            d = shard.peek()['doc_vec'].shape[0]
            index = faiss.index_factory(d, f'HNSW{self.num_links}', faiss.METRIC_INNER_PRODUCT)
            shard_it = more_itertools.chunked(shard, self.batch_size)
            if self.verbose:
                shard_it = pt.tqdm(shard_it, total=math.ceil(self.shard_size/self.batch_size), unit='batch', desc=f'building shard {shardid}')
            for batch in shard_it:
                doc_vecs = np.stack(d['doc_vec'] for d in batch)
                if self.score_fn == 'cos':
                    doc_vecs = doc_vecs / np.linalg.norm(doc_vecs, axis=1, keepdims=True)
                index.add(doc_vecs)
                docnos.extend(d['docno'] for d in batch)
            faiss.write_index(index, str(path/f'{shardid}.faiss'))
        DocnoFile.build(docnos, path/'docnos.npy')













class TorchIndex(NumpyIndex):
    def __init__(self,
                 index_path=None,
                 num_results=1000,
                 score_fn='dot',
                 overwrite=False,
                 batch_size=4096,
                 verbose=False,
                 dtype='f4',
                 drop_query_vec=True,
                 half=False,
                 idx_mem=500000000):
        super().__init__(index_path, num_results, score_fn, overwrite, batch_size, verbose, dtype, drop_query_vec, inmem=False)
        self.half = half
        self.idx_mem = idx_mem
        self._meta = None
        self._cpu_data = None
        self._cuda_data = None
        self._docnos = None
        self._did_start = None
        self._cuda_slice = None

    def docnos_and_data(self):
        if self._meta is None:
            with (self.index_path/'meta.json').open('rt') as f_meta:
                self._meta = json.load(f_meta)
        meta = self._meta
        if self._cpu_data is None:
            SType, TType, CTType, SIZE = {
                'f4': (torch.FloatStorage, torch.FloatTensor, torch.cuda.FloatTensor, 4),
                'f2': (torch.HalfStorage,  torch.HalfTensor,  torch.cuda.HalfTensor,  2),
            }[self.dtype]
            self._cpu_data = TType(SType.from_file(str(self.index_path/'index.npy'), size=meta['count'] * meta['vec_size'])).reshape(meta['count'], meta['vec_size'])
            doc_batch_size = self.idx_mem//SIZE//meta['vec_size']
            if self.half:
                self._cuda_data = torch.cuda.HalfTensor(size=(doc_batch_size, meta['vec_size']), device='cuda')
            else:
                self._cuda_data = CTType(size=(doc_batch_size, meta['vec_size']), device='cuda')
            if self.verbose and doc_batch_size > self.num_results:
                _logger.info(f'TorchIndex using document batches of size {doc_batch_size}')
            if self.num_results >= doc_batch_size:
                _logger.warn(f'TorchIndex using document batches of size {doc_batch_size} but this is less than num_rsults={self.num_results}. Consider using a larger idx_mem (currently {self.idx_mem})')
            self._docnos = DocnoFile(self.index_path/'docnos.npy')

    def transform(self, inp):
        columns = set(inp.columns)
        assert all(f in columns for f in ['qid', 'query_vec']), "TrochIndex expects columns ['qid', 'query_vec'] when used in a pipeline"
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = torch.from_numpy(query_vecs).cuda() # TODO: can this go directly to CUDA? device='cuda' doesn't work
        if self.half:
            query_vecs = query_vecs.half()

        self.docnos_and_data()

        step = self._cuda_data.shape[0]
        it = range(0, self._cpu_data.shape[0], step)
        if self.verbose:
            it = pt.tqdm(it, desc='TorchIndex scoring', unit='docbatch')

        ranked_lists = RankedLists(self.num_results, query_vecs.shape[0])
        for start_idx in it:
            end_idx = start_idx + step
            batch = self._cpu_data[start_idx:end_idx]
            bsize = batch.shape[0]
            if self.half:
                self._cuda_data[:bsize] = batch.half()
            else:
                self._cuda_data[:bsize] = batch

            scores = query_vecs @ self._cuda_data[:bsize].T
            if self.half:
                scores = scores.float()
            if scores.shape[0] > self.num_results:
                scores, dids = torch.topk(scores, k=self.num_results, dim=1)
            else:
                scores, dids = torch.sort(scores, dim=1, descending=False)
            scores = scores.cpu().float().numpy()

            dids = (dids + start_idx).cpu().numpy()
            ranked_lists.update(scores, dids)

        result_scores, result_dids = ranked_lists.results()
        unique_dids, unique_inverse = np.unique(result_dids, return_inverse=True)
        unique_docnos = self._docnos[unique_dids]
        result_docnos = unique_docnos[unique_inverse].reshape(query_vecs.shape[0], -1)
        res = []
        for query, scores, docnos in zip(inp.itertuples(index=False), result_scores, result_docnos):
            if self.drop_query_vec:
                query = query._replace(query_vec=None)
            for score, docno in zip(scores, docnos):
                res.append((*query, docno, score))
        res = pd.DataFrame(res, columns=list(query._fields) + ['docno', 'score'])
        if self.drop_query_vec:
            res = res.drop(columns=['query_vec'])
        res = res[~res.score.isna()]
        res = add_ranks(res)
        return res



__version__ = '0.4.0'

from pyterrier_dr.util import SimFn, infer_device
from pyterrier_dr.indexes import DocnoFile, NilIndex, NumpyIndex, RankedLists, FaissFlat, FaissHnsw, MemIndex, TorchIndex
from pyterrier_dr.flex import FlexIndex
from pyterrier_dr.biencoder import BiEncoder, BiQueryEncoder, BiDocEncoder, BiScorer
from pyterrier_dr.hgf_models import HgfBiEncoder, TasB, RetroMAE
from pyterrier_dr.sbert_models import SBertBiEncoder, Ance, Query2Query, GTR, E5
from pyterrier_dr.tctcolbert_model import TctColBert
from pyterrier_dr.electra import ElectraScorer
from pyterrier_dr.bge_m3 import BGEM3, BGEM3QueryEncoder, BGEM3DocEncoder
from pyterrier_dr.cde import CDE, CDECache
from pyterrier_dr.prf import AveragePrf, VectorPrf

__all__ = ["FlexIndex", "DocnoFile", "NilIndex", "NumpyIndex", "RankedLists", "FaissFlat", "FaissHnsw", "MemIndex", "TorchIndex",
           "BiEncoder", "BiQueryEncoder", "BiDocEncoder", "BiScorer", "HgfBiEncoder", "TasB", "RetroMAE", "SBertBiEncoder", "Ance",
           "Query2Query", "GTR", "E5", "TctColBert", "ElectraScorer", "BGEM3", "BGEM3QueryEncoder", "BGEM3DocEncoder", "CDE", "CDECache",
           "SimFn", "infer_device", "AveragePrf", "VectorPrf"]



import tempfile
import unittest
import numpy as np
import pandas as pd
import pyterrier_dr
from pyterrier_dr import FlexIndex


class TestFlexIndex(unittest.TestCase):

    def _generate_data(self, count=2000, dim=100):
        def random_unit_vec():
            v = np.random.rand(dim).astype(np.float32)
            return v / np.linalg.norm(v)
        return [
            {'docno': str(i), 'doc_vec': random_unit_vec()}
            for i in range(count)
        ]

    def test_index_typical(self):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')

            self.assertFalse(index.built())

            dataset = self._generate_data()

            index.index(dataset)

            self.assertTrue(index.built())

            self.assertEqual(len(index), len(dataset))

            stored_dataset = list(index.get_corpus_iter())
            self.assertEqual(len(stored_dataset), len(dataset))
            for a, b in zip(stored_dataset, dataset):
                self.assertEqual(a['docno'], b['docno'])
                self.assertTrue((a['doc_vec'] == b['doc_vec']).all())

    def test_corpus_graph(self):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data()
            index.index(dataset)
            
            graph = index.corpus_graph(16)
            self.assertEqual(graph.neighbours(4).shape, (16,))

    @unittest.skipIf(not pyterrier_dr.util.faiss_available(), "faiss not available")
    def test_faiss_hnsw_graph(self):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data()
            index.index(dataset)

            graph = index.faiss_hnsw_graph(16)
            self.assertEqual(graph.neighbours(4).shape, (16,))

    def _test_retr(self, Retr, exact=True, test_smaller=True):
        with self.subTest('basic'):
            with tempfile.TemporaryDirectory() as destdir:
                index = FlexIndex(destdir+'/index')
                dataset = self._generate_data(count=2000)
                index.index(dataset)
                
                retr = Retr(index)
                res = retr(pd.DataFrame([
                    {'qid': '0', 'query_vec': dataset[0]['doc_vec']},
                    {'qid': '1', 'query_vec': dataset[1]['doc_vec']},
                ]))
                self.assertTrue(all(c in res.columns) for c in ['qid', 'docno', 'rank', 'score', 'query_vec'])
                if exact:
                    self.assertEqual(len(res), 2000)
                    self.assertEqual(len(res[res.qid=='0']), 1000)
                    self.assertEqual(len(res[res.qid=='1']), 1000)
                    self.assertEqual(res[(res.qid=='0')&((res['rank']==0))].iloc[0]['docno'], '0')
                    self.assertEqual(res[(res.qid=='1')&((res['rank']==0))].iloc[0]['docno'], '1')
                else:
                    self.assertTrue(len(res) <= 2000)
                    self.assertTrue(len(res[res.qid=='0']) <= 1000)
                    self.assertTrue(len(res[res.qid=='1']) <= 1000)

        with self.subTest('drop_query_vec=True'):
            with tempfile.TemporaryDirectory() as destdir:
                index = FlexIndex(destdir+'/index')
                dataset = self._generate_data(count=2000)
                index.index(dataset)
                
                retr = Retr(index, drop_query_vec=True)
                res = retr(pd.DataFrame([
                    {'qid': '0', 'query_vec': dataset[0]['doc_vec']},
                    {'qid': '1', 'query_vec': dataset[1]['doc_vec']},
                ]))
                self.assertTrue(all(c in res.columns) for c in ['qid', 'docno', 'rank', 'score'])
                self.assertTrue(all(c not in res.columns) for c in ['query_vec'])

        if test_smaller:
            with self.subTest('smaller'):
                with tempfile.TemporaryDirectory() as destdir:
                    index = FlexIndex(destdir+'/index')
                    dataset = self._generate_data(count=100)
                    index.index(dataset)
                    
                    retr = Retr(index)
                    res = retr(pd.DataFrame([
                        {'qid': '0', 'query_vec': dataset[0]['doc_vec']},
                        {'qid': '1', 'query_vec': dataset[1]['doc_vec']},
                    ]))
                    self.assertTrue(all(c in res.columns) for c in ['qid', 'docno', 'rank', 'score'])
                    if exact:
                        self.assertEqual(len(res), 200)
                        self.assertEqual(len(res[res.qid=='0']), 100)
                        self.assertEqual(len(res[res.qid=='1']), 100)
                        self.assertEqual(res[(res.qid=='0')&((res['rank']==0))].iloc[0]['docno'], '0')
                        self.assertEqual(res[(res.qid=='1')&((res['rank']==0))].iloc[0]['docno'], '1')
                    else:
                        self.assertTrue(len(res) <= 200)
                        self.assertTrue(len(res[res.qid=='0']) <= 100)
                        self.assertTrue(len(res[res.qid=='1']) <= 100)

    @unittest.skipIf(not pyterrier_dr.util.faiss_available(), "faiss not available")
    def test_faiss_flat_retriever(self):
        self._test_retr(FlexIndex.faiss_flat_retriever)

    @unittest.skipIf(not pyterrier_dr.util.faiss_available(), "faiss not available")
    def test_faiss_hnsw_retriever(self):
        self._test_retr(FlexIndex.faiss_hnsw_retriever, exact=False)

    @unittest.skipIf(not pyterrier_dr.util.faiss_available(), "faiss not available")
    def test_faiss_ivf_retriever(self):
        self._test_retr(FlexIndex.faiss_ivf_retriever, exact=False)

    @unittest.skipIf(not pyterrier_dr.util.scann_available(), "scann not available")
    def test_scann_retriever(self):
        self._test_retr(FlexIndex.scann_retriever, exact=False)

    @unittest.skipIf(not pyterrier_dr.util.flatnav_available(), "flatnav not available")
    def test_flatnav_retriever(self):
        self._test_retr(FlexIndex.flatnav_retriever, exact=False)

    def test_np_retriever(self):
        self._test_retr(FlexIndex.np_retriever)

    def test_torch_retriever(self):
        self._test_retr(FlexIndex.torch_retriever)

    @unittest.skipIf(not pyterrier_dr.util.voyager_available(), "voyager not available")
    def test_voyager_retriever(self):
        # Voyager doesn't support requesting more results than are availabe in the index
        # (the "smaller" case), so disable that test case here.
        self._test_retr(FlexIndex.voyager_retriever, exact=False, test_smaller=False)

    def test_np_vec_loader(self):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data()
            index.index(dataset)
            
            vec_loader = index.np_vec_loader()
            with self.subTest('docid'):
                res = vec_loader(pd.DataFrame({
                    'docid': [5, 1, 100, 198],
                }))
                self.assertTrue(all(c in res.columns) for c in ['docid', 'doc_vec'])
                self.assertEqual(len(res), 4)
                self.assertTrue((res.iloc[0]['doc_vec'] == dataset[5]['doc_vec']).all())
                self.assertTrue((res.iloc[1]['doc_vec'] == dataset[1]['doc_vec']).all())
                self.assertTrue((res.iloc[2]['doc_vec'] == dataset[100]['doc_vec']).all())
                self.assertTrue((res.iloc[3]['doc_vec'] == dataset[198]['doc_vec']).all())
            with self.subTest('docno'):
                res = vec_loader(pd.DataFrame({
                    'docno': ['20', '0', '100', '198'],
                    'query': 'ignored',
                }))
                self.assertTrue(all(c in res.columns) for c in ['docno', 'doc_vec', 'query'])
                self.assertEqual(len(res), 4)
                self.assertTrue((res.iloc[0]['doc_vec'] == dataset[20]['doc_vec']).all())
                self.assertTrue((res.iloc[1]['doc_vec'] == dataset[0]['doc_vec']).all())
                self.assertTrue((res.iloc[2]['doc_vec'] == dataset[100]['doc_vec']).all())
                self.assertTrue((res.iloc[3]['doc_vec'] == dataset[198]['doc_vec']).all())

    def _test_reranker(self, Reranker, expands=False):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data(count=2000)
            index.index(dataset)
            
            retr = Reranker(index)
            res = retr(pd.DataFrame([
                {'qid': '0', 'query_vec': dataset[0]['doc_vec'], 'docno': '0', 'score': -100.},
                {'qid': '0', 'query_vec': dataset[0]['doc_vec'], 'docno': '50', 'score': -100.},
                {'qid': '1', 'query_vec': dataset[1]['doc_vec'], 'docno': '100', 'score': -100.},
                {'qid': '1', 'query_vec': dataset[1]['doc_vec'], 'docno': '0', 'score': -100.},
                {'qid': '1', 'query_vec': dataset[1]['doc_vec'], 'docno': '40', 'score': -100.},
            ]))
            self.assertTrue(all(c in res.columns) for c in ['qid', 'docno', 'rank', 'score'])
            if expands:
                self.assertTrue(len(res) >= 5)
            else:
                self.assertEqual(len(res), 5)
            self.assertEqual(res[(res.qid=='0')&((res['rank']==0))].iloc[0]['docno'], '0')
            self.assertEqual(len(res[res.score==-100.]), 0) # all scores re-assigned

    def test_np_scorer(self):
        self._test_reranker(FlexIndex.np_scorer)

    def test_torch_scorer(self):
        self._test_reranker(FlexIndex.torch_scorer)

    def test_pre_ladr(self):
        self._test_reranker(FlexIndex.pre_ladr, expands=True)

    def test_ada_ladr(self):
        self._test_reranker(FlexIndex.ada_ladr, expands=True)

    def test_gar(self):
        self._test_reranker(FlexIndex.gar, expands=True)

    def test_torch_vecs(self):
        with tempfile.TemporaryDirectory() as destdir:
            index = FlexIndex(destdir+'/index')
            dataset = self._generate_data()
            index.index(dataset)
            
            torch_vecs = index.torch_vecs()
            self.assertEqual(torch_vecs.shape, (2000, 100))


if __name__ == '__main__':
    unittest.main()



from more_itertools import chunked
import numpy as np
import torch
from transformers import AutoTokenizer, AutoModel
from pyterrier_dr.util import Variants
from . import BiEncoder


class TctColBert(BiEncoder, metaclass=Variants):
    """Dense encoder for TCT-ColBERT (Tightly-Coupled Teachers over ColBERT)

    See :class:`~pyterrier_dr.BiEncoder` for usage information.

    .. cite.dblp:: journals/corr/abs-2010-11386

    .. automethod:: base()
    .. automethod:: hn()
    .. automethod:: hnp()
    """
    def __init__(self, model_name=None, batch_size=32, text_field='text', verbose=False, device=None):
        super().__init__(batch_size=batch_size, text_field=text_field, verbose=verbose)
        self.model_name = model_name or next(iter(self.VARIANTS.values()))
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.model = AutoModel.from_pretrained(self.model_name).to(self.device).eval()

    def encode_queries(self, texts, batch_size=None):
        results = []
        with torch.no_grad():
            for chunk in chunked(texts, batch_size or self.batch_size):
                inps = self.tokenizer([f'[CLS] [Q] {q} ' + ' '.join(['[MASK]'] * 32) for q in chunk], add_special_tokens=False, return_tensors='pt', padding=True, truncation=True, max_length=36)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                res = self.model(**inps).last_hidden_state
                res = res[:, 4:, :].mean(dim=1) # remove the first 4 tokens (representing [CLS] [ Q ]), and average
                results.append(res.cpu().numpy())
        if not results:
            return np.empty(shape=(0, 0))
        return np.concatenate(results, axis=0)

    def encode_docs(self, texts, batch_size=None):
        results = []
        with torch.no_grad():
            for chunk in chunked(texts, batch_size or self.batch_size):
                inps = self.tokenizer([f'[CLS] [D] {d}' for d in chunk], add_special_tokens=False, return_tensors='pt', padding=True, truncation=True, max_length=512)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                res = self.model(**inps).last_hidden_state
                res = res[:, 4:, :] # remove the first 4 tokens (representing [CLS] [ D ])
                res = res * inps['attention_mask'][:, 4:].unsqueeze(2) # apply attention mask
                lens = inps['attention_mask'][:, 4:].sum(dim=1).unsqueeze(1)
                lens[lens == 0] = 1 # avoid edge case of div0 errors
                res = res.sum(dim=1) / lens # average based on dim
                results.append(res.cpu().numpy())
        if not results:
            return np.empty(shape=(0, 0))
        return np.concatenate(results, axis=0)

    def __repr__(self):
        return f'TctColBert({repr(self.model_name)})'

    VARIANTS = {
        'base': 'castorini/tct_colbert-msmarco',
        'hn': 'castorini/tct_colbert-v2-hn-msmarco',
        'hnp': 'castorini/tct_colbert-v2-hnp-msmarco',
    }



from typing import Union, Iterable, Dict
import shutil
import itertools
import json
from pathlib import Path
import numpy as np
import more_itertools
import pandas as pd
import pyterrier as pt
from npids import Lookup
from enum import Enum
from .. import SimFn
import pyterrier_alpha as pta


class IndexingMode(Enum):
    create = "create"
    overwrite = "overwrite"
    # append???


class FlexIndex(pta.Artifact, pt.Indexer):
    """ Represents a FLexible EXecution (FLEX) Index, which is a dense index format.

    FLEX allows for a variety of retrieval implementations (NumPy, FAISS, etc.) and algorithms (exhaustive, HNSW, etc.)
    to be tested. In most cases, the same vector storage can be used across implementations and algorithms, saving
    considerably on disk space.
    """

    ARTIFACT_TYPE = 'dense_index'
    ARTIFACT_FORMAT = 'flex'


    def __init__(self,
        path: str,
        *,
        sim_fn: Union[SimFn, str] = SimFn.dot,
        verbose: bool = True
    ):
        """
        Args:
            path: The path to the index directory
            sim_fn: The similarity function to use
            verbose: Whether to display verbose output (e.g., progress bars)
        """
        super().__init__(path)
        self.index_path = Path(path)
        self.sim_fn = SimFn(sim_fn)
        self.verbose = verbose
        self._meta = None
        self._docnos = None
        self._dvecs = None
        self._cache = {}

    def payload(self, return_dvecs=True, return_docnos=True):
        if self._meta is None:
            with open(self.index_path/'pt_meta.json', 'rt') as f_meta:
                meta = json.load(f_meta)
                assert meta.get('type') == 'dense_index' and meta['format'] == 'flex'
                self._meta = meta
        res = [self._meta]
        if return_dvecs:
            if self._dvecs is None:
                self._dvecs = np.memmap(self.index_path/'vecs.f4', mode='r', dtype=np.float32, shape=(self._meta['doc_count'], self._meta['vec_size']))
            res.insert(0, self._dvecs)
        if return_docnos:
            if self._docnos is None:
                self._docnos = Lookup(self.index_path/'docnos.npids')
            res.insert(0, self._docnos)
        return res

    def __len__(self):
        meta, = self.payload(return_dvecs=False, return_docnos=False)
        return meta['doc_count']

    def index(self, inp: Iterable[Dict]) -> pta.Artifact:
        """Index the given input data stream to a new index at this location.

        Each record in ``inp`` is expected to be a dictionary containing at least two keys: ``docno`` (a unique document
        identifier) and ``doc_vec`` (a dense vector representation of the document).

        Typically this method will be used in a pipeline of operations, where the input data is first transformed by a
        document encoder to add the ``doc_vec`` values before it is indexed. For example:

        .. code-block:: python
            :caption: Index documents into a :class:`~pyterrier_dr.FlexIndex` using a :class:`~pyterrier_dr.TasB` encoder.
            
            from pyterrier_dr import TasB, FlexIndex    
            encoder = TasB.dot()
            index = FlexIndex('my_index')
            pipeline = encoder >> index
            pipeline.index([
                {'docno': 'doc1', 'text': 'hello'},
                {'docno': 'doc2', 'text': 'world'},
            ])

        Args:
            inp: An iterable of dictionaries to index.

        Returns:
            :class:`pyterrier_alpha.Artifact`: A reference back to this index (``self``).

        Raises:
            RuntimeError: If the index is aready built.
        """
        return self.indexer().index(inp)

    def indexer(self, *, mode: Union[IndexingMode, str] = IndexingMode.create) -> 'FlexIndexer':
        """Return an indexer for this index with the specified options.

        This transformer gives more fine-grained control over the indexing process, allowing you to specify whether
        to create a new index or overwrite an existing one.

        Similar to :meth:`index`, this method will typically be used in a pipeline of operations, where the input data
        is first transformed by a document encoder to add the ``doc_vec`` values before it is indexed. For example:

        .. code-block:: python
            :caption: Oerwrite a :class:`~pyterrier_dr.FlexIndex` using a :class:`~pyterrier_dr.TasB` encoder.
            
            from pyterrier_dr import TasB, FlexIndex    
            encoder = TasB.dot()
            index = FlexIndex('my_index')
            pipeline = encoder >> index.indexer(mode='overwrite')
            pipeline.index([
                {'docno': 'doc1', 'text': 'hello'},
                {'docno': 'doc2', 'text': 'world'},
            ])

        Args:
            mode: The indexing mode to use (``create`` or ``overwrite``).

        Returns:
            :class:`~pyterrier.Indexer`: A new indexer instance.
        """
        return FlexIndexer(self, mode=mode)

    def transform(self, inp):
        with pta.validate.any(inp) as v:
            v.query_frame(extra_columns=['query_vec'], mode='retriever')
            v.result_frame(extra_columns=['query_vec'], mode='scorer')

        if v.mode == 'retriever':
            return self.retriever()(inp)
        if v.mode == 'scorer':
            return self.scorer()(inp)

    def get_corpus_iter(self, start_idx=None, stop_idx=None, verbose=True) -> Iterable[Dict]:
        """Iterate over the documents in the index.

        Args:
            start_idx: The index of the first document to return (or ``None`` to start at the first document).
            stop_idx: The index of the last document to return (or ``None`` to end on the last document).
            verbose: Whether to display a progress bar.

        Yields:
            Dict[str,Any]: A dictionary with keys ``docno`` and ``doc_vec``.
        """
        docnos, dvecs, meta = self.payload()
        docno_iter = iter(docnos)
        if start_idx is not None or stop_idx is not None:
            docno_iter = itertools.islice(docno_iter, start_idx, stop_idx)
            dvecs = dvecs[start_idx:stop_idx]
        it = zip(docno_iter, range(dvecs.shape[0]))
        if self.verbose:
            it = pt.tqdm(it, total=dvecs.shape[0], desc=f'{str(self.index_path)} dvecs')
        for docno, i in it:
            yield {'docno': docno, 'doc_vec': dvecs[i]}

    def _load_docids(self, inp):
        with pta.validate.any(inp) as v:
            v.columns(includes=['docid'], mode='docid')
            v.columns(includes=['docno'], mode='docno')
        if v.mode == 'docid':
            return inp['docid'].values
        docnos, config = self.payload(return_dvecs=False)
        return docnos.inv[inp['docno'].values] # look up docids from docnos

    def built(self) -> bool:
        """Check if the index has been built.

        Returns:
            bool: ``True`` if the index has been built, otherwise ``False``.
        """
        return self.index_path.exists()

    def docnos(self) -> Lookup:
        """Return the document identifier (docno) lookup data structure.

        Returns:
            :class:`npids.Lookup`: The document number lookup.
        """
        docnos, meta = self.payload(return_dvecs=False)
        return docnos


class FlexIndexer(pt.Indexer):
    def __init__(self, index: FlexIndex, mode: Union[IndexingMode, str] = IndexingMode.create):
        self._index = index
        self.mode = IndexingMode(mode)

    def __repr__(self):
        return f'{self._index}.indexer(mode={self.mode!r})'

    def transform(self, inp):
        raise RuntimeError("FlexIndexer cannot be used as a transformer, use .index() instead")

    def index(self, inp):
        if isinstance(inp, pd.DataFrame):
            inp = inp.to_dict(orient="records")
        inp = more_itertools.peekable(inp)
        path = Path(self._index.index_path)
        if path.exists():
            if self.mode == IndexingMode.overwrite:
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Index already exists at {self._index.index_path}. If you want to delete and re-create an existing index, you can pass index.indexer(mode="overwrite")')
        path.mkdir(parents=True, exist_ok=True)
        vec_size = None
        count = 0
        if self._index.verbose:
            inp = pt.tqdm(inp, desc='indexing', unit='dvec')
        with open(path/'vecs.f4', 'wb') as fout, Lookup.builder(path/'docnos.npids') as docnos:
            for d in inp:
                vec = d['doc_vec']
                if vec_size is None:
                    vec_size = vec.shape[0]
                elif vec_size != vec.shape[0]:
                    raise ValueError(f'Inconsistent vector shapes detected (expected {vec_size} but found {vec.shape[0]})')
                vec = vec.astype(np.float32)
                fout.write(vec.tobytes())
                docnos.add(d['docno'])
                count += 1
        with open(path/'pt_meta.json', 'wt') as f_meta:
            json.dump({
                "type": self._index.ARTIFACT_TYPE,
                "format": self._index.ARTIFACT_FORMAT,
                "vec_size": vec_size,
                "doc_count": count
            }, f_meta)
        return self._index


def _load_dvecs(flex_index, inp):
    dvecs, config = flex_index.payload(return_docnos=False)
    return dvecs[flex_index._load_docids(inp)]



from enum import Enum
import torch


class SimFn(Enum):
    dot = 'dot'
    cos = 'cos'


class Variants(type):
    def __getattr__(cls, name):
        if name in cls.VARIANTS:
            @staticmethod
            def wrapped(*args, **kwargs):
                return cls(cls.VARIANTS[name], *args, **kwargs)
            wrapped = wrapped.__get__(cls)
            wrapped.__doc__ = f"Model: ``{cls.VARIANTS[name]}`` `[link] <https://huggingface.co/{cls.VARIANTS[name]}>`__"
            if name == next(iter(cls.VARIANTS)):
                wrapped.__doc__ = '*(default)* ' + wrapped.__doc__
            return wrapped

    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs)


def infer_device(device=None):
    if device is None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
    return torch.device(device)


def package_available(name):
    try:
        __import__(name)
        return True
    except ImportError:
        return False


def faiss_available():
    return package_available('faiss')


def assert_faiss():
    assert faiss_available(), "faiss required; install using instructions here: <https://github.com/facebookresearch/faiss/blob/main/INSTALL.md>"


def scann_available():
    if not package_available('scann'):
        return False
    import scann
    # version 1.0.0 is requiredl detect via scann.scann_ops_pybind.builder
    return hasattr(scann.scann_ops_pybind, 'builder')


def assert_scann():
    assert scann_available(), "scann==1.0.0 required; install from wheel here: <https://github.com/google-research/google-research/blob/master/scann/docs/releases.md#scann-wheel-archive>"


def voyager_available():
    return package_available('voyager')


def assert_voyager():
    assert voyager_available(), "voyager required; install with `pip install voyager`"


def flatnav_available():
    return package_available('flatnav')


def assert_flatnav():
    assert flatnav_available(), "flatnav required; install with instructions here: https://github.com/BlaiseMuhirwa/flatnav"



import pyterrier as pt
import heapq
import pyterrier_alpha as pta
from . import FlexIndex
import numpy as np
from pyterrier_dr import SimFn

class FlexGar(pt.Transformer):
    def __init__(self, flex_index, graph, score_fn, batch_size=128, num_results=1000, drop_query_vec=False):
        self.flex_index = flex_index
        self.docnos, self.dvecs, _ = flex_index.payload()
        self.score_fn = score_fn
        self.graph = graph
        self.batch_size = batch_size
        self.num_results = num_results
        self.drop_query_vec = drop_query_vec

    def transform(self, inp):
        pta.validate.result_frame(inp, extra_columns=['query_vec', 'score'])

        qcols = [col for col in inp.columns if col.startswith('q') and col != 'query_vec']
        if not self.drop_query_vec:
            qcols += ['query_vec']
        all_results = pta.DataFrameBuilder(qcols + ['docno', 'score', 'rank'])

        for qid, inp in inp.groupby('qid'):
            qvec = inp['query_vec'].iloc[0].reshape(1, -1)
            qdata = {col: [inp[col].iloc[0]] for col in qcols}
            initial_heap = list(zip(-inp['score'], self.docnos.inv[inp['docno']]))
            heapq.heapify(initial_heap)
            results = {}
            frontier_heap = []
            frontier_leftover = []
            i = 0
            while len(results) < self.num_results and (initial_heap or frontier_heap or frontier_leftover):
                size = min(self.batch_size, self.num_results - len(results))
                batch = []
                if i % 2 == 0 and initial_heap:
                    while len(batch) < size and initial_heap:
                        did = heapq.heappop(initial_heap)[1]
                        if did not in results:
                            batch.append(did)
                elif i % 2 == 1 and (frontier_heap or frontier_leftover):
                    while len(batch) < size and (frontier_heap or frontier_leftover):
                        if frontier_leftover:
                            frontier_leftover, e = frontier_leftover[size-len(batch):], frontier_leftover[:size-len(batch)]
                            batch.extend(e)
                        else:
                            _, did = heapq.heappop(frontier_heap)
                            frontier_leftover.extend(d for d in self.graph.neighbours(did) if d not in results)
                dvecs = self.dvecs[np.array(batch).astype(int)]
                if self.score_fn == SimFn.dot:
                    scores = (qvec @ dvecs.T).reshape(-1)
                for did, score in zip(batch, scores):
                    results[did] = score
                    heapq.heappush(frontier_heap, (-score, did))
                i += 1
            d, s = zip(*sorted(results.items(), key=lambda x: (-x[1], x[0])))
            all_results.extend(dict(
                **qdata,
                docno=self.docnos.fwd[d],
                score=s,
                rank=np.arange(len(s)),
            ))
        return all_results.to_df()


def _gar(self,
    k: int = 16,
    *,
    batch_size: int = 128,
    num_results: int = 1000
) -> pt.Transformer:
    """Returns a retriever that uses a corpus graph to search over a FlexIndex.

    Args:
        k (int): Number of neighbours in the corpus graph. Defaults to 16.
        batch_size (int): Batch size for retrieval. Defaults to 128.
        num_results (int): Number of results per query to return. Defaults to 1000.

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses a corpus graph to search over a FlexIndex.

    .. cite.dblp:: conf/cikm/MacAvaneyTM22
    """
    return FlexGar(self, self.corpus_graph(k), SimFn.dot, batch_size=batch_size, num_results=num_results)
FlexIndex.gar = _gar



from typing import Optional
import json
import math
import struct
import os
import pyterrier as pt
import numpy as np
import tempfile
import ir_datasets
import pyterrier_dr
import pyterrier_alpha as pta
from . import FlexIndex

logger = ir_datasets.log.easy()


class FaissRetriever(pt.Indexer):
    def __init__(self, flex_index, faiss_index, n_probe=None, ef_search=None, search_bounded_queue=None, qbatch=64, drop_query_vec=False, num_results=1000):
        self.flex_index = flex_index
        self.faiss_index = faiss_index
        self.n_probe = n_probe
        self.ef_search = ef_search
        self.search_bounded_queue = search_bounded_queue
        self.qbatch = qbatch
        self.drop_query_vec = drop_query_vec
        self.num_results = num_results

    def transform(self, inp):
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        docnos, config = self.flex_index.payload(return_dvecs=False)
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = query_vecs.copy()
        num_q = query_vecs.shape[0]
        QBATCH = self.qbatch
        if self.n_probe is not None:
            self.faiss_index.nprobe = self.n_probe
        if self.ef_search is not None:
            self.faiss_index.hnsw.efSearch = self.ef_search
        if self.search_bounded_queue is not None:
            self.faiss_index.hnsw.search_bounded_queue = self.search_bounded_queue
        it = range(0, num_q, QBATCH)
        if self.flex_index.verbose:
            it = pt.tqdm(it, unit='qbatch')

        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        for qidx in it:
            scores, dids = self.faiss_index.search(query_vecs[qidx:qidx+QBATCH], self.num_results)
            for s, d in zip(scores, dids):
                mask = d != -1
                d = d[mask]
                s = s[mask]
                result.extend({
                    'docno': docnos.fwd[d],
                    'docid': d,
                    'score': s,
                    'rank': np.arange(d.shape[0]),
                })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


def _faiss_flat_retriever(self, *, gpu=False, qbatch=64, drop_query_vec=False):
    """Returns a retriever that uses FAISS to perform brute-force search over the indexed vectors.

    Args:
        gpu: Whether to load the index onto GPU for scoring
        qbatch: The batch size during search
        drop_query_vec: Whether to drop the query vector from the output

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses FAISS to perform brute-force search over the indexed vectors

    .. note::
        This transformer requires the ``faiss`` package to be installed.

    .. cite.dblp:: journals/corr/abs-2401-08281
    """
    pyterrier_dr.util.assert_faiss()
    import faiss
    if 'faiss_flat' not in self._cache:
        meta, = self.payload(return_dvecs=False, return_docnos=False)
        with tempfile.TemporaryDirectory() as tmp, logger.duration('reading faiss flat'):
            DUMMY = b'\x00\x00\x10\x00\x00\x00\x00\x00'
            #       [ type ]                                                                             [ train ]  [ metric (dot)     ]
            header = b'IxFI'  + struct.pack('<IQ', meta['vec_size'], meta['doc_count']) + DUMMY + DUMMY  + b'\x00' + b'\x00\x00\x00\x00' + struct.pack('<Q', meta['doc_count'] * meta['vec_size'])
            with open(os.path.join(tmp, 'tmp'), 'wb') as fout:
                fout.write(b' ' + header)
            reader = faiss.BufferedIOReader(faiss.FileIOReader(os.path.join(tmp, 'tmp')))
            reader.read_bytes(1)
            reader.reader = faiss.FileIOReader(str(self.index_path/'vecs.f4'))
            self._cache['faiss_flat'] = faiss.read_index(reader)
    if gpu:
        if 'faiss_flat_gpu' not in self._cache:
            co = faiss.GpuMultipleClonerOptions()
            co.shard = True
            self._cache['faiss_flat_gpu'] = faiss.index_cpu_to_all_gpus(self._faiss_flat, co=co)
        return FaissRetriever(self, self._cache['faiss_flat_gpu'], drop_query_vec=drop_query_vec)
    return FaissRetriever(self, self._cache['faiss_flat'], qbatch=qbatch, drop_query_vec=drop_query_vec)
FlexIndex.faiss_flat_retriever = _faiss_flat_retriever


def _faiss_hnsw_retriever(
    self,
    neighbours: int = 32,
    *,
    num_results: int = 1000,
    ef_construction: int = 40,
    ef_search: int = 16,
    cache: bool = True,
    search_bounded_queue: bool = True,
    qbatch: int = 64,
    drop_query_vec: bool = False,
) -> pt.Transformer:
    """Returns a retriever that uses FAISS over a HNSW index.

    Creates the HNSW graph structure if it does not already exist. When ``cache=True`` (dfault), this graph structure is
    cached to disk for subsequent use.

    Args:
        neighbours: The number of neighbours of the constructed neighborhood graph
        num_results: The number of results to return per query
        ef_construction: The number of neighbours to consider during construction
        ef_search: The number of neighbours to consider during search
        cache: Whether to cache the index to disk
        search_bounded_queue: Whether to use a bounded queue during search
        qbatch: The batch size during search
        drop_query_vec: Whether to drop the query vector from the output

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses FAISS over a HNSW index

    .. note::
        This transformer requires the ``faiss`` package to be installed.

    .. cite.dblp:: journals/corr/abs-2401-08281
    """
    pyterrier_dr.util.assert_faiss()
    import faiss
    meta, = self.payload(return_dvecs=False, return_docnos=False)

    key = ('faiss_hnsw', neighbours, ef_construction)
    index_name = f'hnsw_n-{neighbours}_ef-{ef_construction}.faiss'
    if key not in self._cache:
        dvecs, meta = self.payload(return_docnos=False)
        if not os.path.exists(self.index_path/index_name):
            idx = faiss.IndexHNSWFlat(meta['vec_size'], neighbours, faiss.METRIC_INNER_PRODUCT)
            for start_idx in pt.tqdm(range(0, dvecs.shape[0], 4096), desc='indexing', unit='batch'):
                idx.add(np.array(dvecs[start_idx:start_idx+4096]))
            idx.storage = faiss.IndexFlatIP(meta['vec_size']) # clear storage ; we can use faiss_flat here instead so we don't keep an extra copy
            if cache:
                with logger.duration('caching index'):
                    faiss.write_index(idx, str(self.index_path/index_name))
            self._cache[key] = idx
        else:
            with logger.duration('reading hnsw table'):
                self._cache[key] = faiss.read_index(str(self.index_path/index_name))
        self._cache[key].storage = self.faiss_flat_retriever().faiss_index
    return FaissRetriever(self, self._cache[key], num_results=num_results, ef_search=ef_search, search_bounded_queue=search_bounded_queue, qbatch=qbatch, drop_query_vec=drop_query_vec)
FlexIndex.faiss_hnsw_retriever = _faiss_hnsw_retriever


def _faiss_hnsw_graph(self, neighbours: int = 32, *, ef_construction: int = 40):
    """Returns the (approximate) HNSW graph structure created by the HNSW index.

    If the graph structure does not already exist, it is created and cached to disk.

    Args:
        neighbours: The number of neighbours of the constructed neighborhood graph
        ef_construction: The number of neighbours to consider during construction

    Returns:
        :class:`pyterrier_adaptive.CorpusGraph`: The HNSW graph structure

    .. note::
        This function requires the ``faiss`` package to be installed.

    .. cite.dblp:: journals/corr/abs-2401-08281
    """
    key = ('faiss_hnsw', neighbours//2, ef_construction)
    graph_name = f'hnsw_n-{neighbours}_ef-{ef_construction}.graph'
    if key not in self._cache:
        if not (self.index_path/graph_name/'pt_meta.json').exists():
            retr = self.faiss_hnsw_retriever(neighbours=neighbours//2, ef_construction=ef_construction)
            _build_hnsw_graph(retr.faiss_index.hnsw, self.index_path/graph_name)
        from pyterrier_adaptive import CorpusGraph
        self._cache[key] = CorpusGraph.load(self.index_path/graph_name)
    return self._cache[key]
FlexIndex.faiss_hnsw_graph = _faiss_hnsw_graph


def _build_hnsw_graph(hnsw, out_dir):
    lvl_0_size = hnsw.nb_neighbors(0)
    num_docs = hnsw.offsets.size() - 1
    scores = np.zeros(lvl_0_size, dtype=np.float16)
    out_dir.mkdir(parents=True, exist_ok=True)
    edges_path = out_dir/'edges.u32.np'
    weights_path = out_dir/'weights.f16.np'
    with ir_datasets.util.finialized_file(str(edges_path), 'wb') as fe, \
         ir_datasets.util.finialized_file(str(weights_path), 'wb') as fw:
        for did in pt.tqdm(range(num_docs), unit='doc', smoothing=1):
            start = hnsw.offsets.at(did)
            dids = [hnsw.neighbors.at(i) for i in range(start, start+lvl_0_size)]
            dids = [(d if d != -1 else did) for d in dids] # replace with self if missing value
            fe.write(np.array(dids, dtype=np.uint32).tobytes())
            fw.write(scores.tobytes())
    (out_dir/'docnos.npids').symlink_to('../docnos.npids')
    with (out_dir/'pt_meta.json').open('wt') as fout:
        json.dump({
            'type': 'corpus_graph',
            'format': 'np_topk',
            'doc_count': num_docs,
            'k': lvl_0_size,
        }, fout)

def _sample_train(index, count=None):
    dvecs, meta = index.payload(return_docnos=False)
    count = min(count or 10_000, dvecs.shape[0])
    idxs = np.random.RandomState(0).choice(dvecs.shape[0], size=count, replace=False)
    return dvecs[idxs]

def _faiss_ivf_retriever(self,
    *,
    num_results: int = 1000,
    train_sample: Optional[int] = None,
    n_list: Optional[int] = None,
    cache: bool = True,
    n_probe: int = 1,
    drop_query_vec: bool = False
):
    """Returns a retriever that uses FAISS over an IVF index.

    If the IVF structure does not already exist, it is created and cached to disk (when ``cache=True`` (default)).

    Args:
        num_results: The number of results to return per query
        train_sample: The number of training samples to use for training the index. If not provided, a default value is used (approximately the square root of the number of documents).
        n_list: The number of posting lists to use for the index. If not provided, a default value is used (approximately ``train_sample/39``).
        cache: Whether to cache the index to disk.
        n_probe: The number of posting lists to probe during search. The higher the value, the better the approximation will be, but the longer it will take.
        drop_query_vec: Whether to drop the query vector from the output.

    Returns:
        :class:`~pyterrier.Transformer`: A retriever that uses FAISS over an IVF index

    .. note::
        This transformer requires the ``faiss`` package to be installed.

    .. cite.dblp:: journals/corr/abs-2401-08281
    """
    pyterrier_dr.util.assert_faiss()
    import faiss
    meta, = self.payload(return_dvecs=False, return_docnos=False)

    if n_list is None:
        if train_sample is None:
            n_list = math.ceil(math.sqrt(meta['doc_count']))
            # we'll shift it to the nearest power of 2
            n_list = int(1 << math.ceil(math.log2(n_list)))
        else:
            n_list = math.floor(train_sample / 39)
        n_list = max(n_list, 4)

    if train_sample is None:
        train_sample = n_list * 39
    elif 0 < train_sample < 1:
        train_sample = math.ceil(train_sample * meta['doc_count'])

    key = ('faiss_ivf', n_list, train_sample)
    index_name = f'ivf_nlist-{n_list}_train-{train_sample}.faiss'
    if key not in self._cache:
        dvecs, meta = self.payload(return_docnos=False)
        if not os.path.exists(self.index_path/index_name):
            quantizer = faiss.IndexFlatIP(meta['vec_size'])
            if pyterrier_dr.util.infer_device().type == 'cuda':
                quantizer = faiss.index_cpu_to_all_gpus(quantizer)
            idx = faiss.IndexIVFFlat(quantizer, meta['vec_size'], n_list, faiss.METRIC_INNER_PRODUCT)
            with logger.duration(f'loading {train_sample} train samples'):
                train = _sample_train(self, train_sample)
            with logger.duration(f'training ivf with {n_list} posting lists'):
                idx.train(train)
            for start_idx in pt.tqdm(range(0, dvecs.shape[0], 4096), desc='indexing', unit='batch'):
                idx.add(np.array(dvecs[start_idx:start_idx+4096]))
            if cache:
                with logger.duration('caching index'):
                    if pyterrier_dr.util.infer_device().type == 'cuda':
                        idx.quantizer = faiss.index_gpu_to_cpu(idx.quantizer)
                    faiss.write_index(idx, str(self.index_path/index_name))
            self._cache[key] = idx
        else:
            with logger.duration('reading index'):
                self._cache[key] = faiss.read_index(str(self.index_path/index_name))
    return FaissRetriever(self, self._cache[key], num_results=num_results, n_probe=n_probe, drop_query_vec=drop_query_vec)
FlexIndex.faiss_ivf_retriever = _faiss_ivf_retriever



from typing import Optional
import math
import os
import pyterrier as pt
import numpy as np
import ir_datasets
import pyterrier_alpha as pta
import pyterrier_dr
from . import FlexIndex

logger = ir_datasets.log.easy()


class ScannRetriever(pt.Indexer):
    def __init__(self, flex_index, scann_index, num_results=1000, leaves_to_search=None, qbatch=64, drop_query_vec=False):
        self.flex_index = flex_index
        self.scann_index = scann_index
        self.leaves_to_search = leaves_to_search
        self.num_results = num_results
        self.qbatch = qbatch
        self.drop_query_vec = drop_query_vec

    def transform(self, inp):
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        docnos, config = self.flex_index.payload(return_dvecs=False)
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = query_vecs.copy()

        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        num_q = query_vecs.shape[0]
        QBATCH = self.qbatch
        for qidx in range(0, num_q, QBATCH):
            dids, scores = self.scann_index.search_batched(query_vecs[qidx:qidx+QBATCH], leaves_to_search=self.leaves_to_search, final_num_neighbors=self.num_results)
            for s, d in zip(scores, dids):
                mask = d != -1
                d = d[mask]
                s = s[mask]
                result.extend({
                    'docno': docnos.fwd[d],
                    'docid': d,
                    'score': s,
                    'rank': np.arange(d.shape[0]),
                })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


def _scann_retriever(self,
    *,
    n_leaves: Optional[int] = None,
    leaves_to_search: int = 1,
    num_results: int = 1000,
    train_sample: Optional[int] = None,
    drop_query_vec=False
):
    """Returns a retriever over a ScaNN (Scalable Nearest Neighbors) index.

    Args:
        n_leaves (int, optional): Number of leaves in the ScaNN index. Defaults to approximatley sqrt(doc_count).
        leaves_to_search (int, optional): Number of leaves to search. Defaults to 1. The higher the value, the more accurate the search.
        num_results (int, optional): Number of results to return. Defaults to 1000.
        train_sample (int, optional): Number of training samples. Defaults to ``n_leaves*39``.
        drop_query_vec (bool, optional): Whether to drop the query vector from the output.

    Returns:
        :class:`~pyterrier.Transformer`: A transformer that retrieves using ScaNN.

    .. note::
        This method requires the ``scann`` package. Install it via ``pip install scann``.

    .. cite.dblp:: conf/icml/GuoSLGSCK20
    """
    pyterrier_dr.util.assert_scann()
    import scann
    dvecs, meta, = self.payload(return_docnos=False)

    if n_leaves is None:
        # rule of thumb: sqrt(doc_count) (from <https://github.com/google-research/google-research/blob/master/scann/docs/algorithms.md>)
        n_leaves = math.ceil(math.sqrt(meta['doc_count']))
        # we'll shift it to the nearest power of 2
        n_leaves = int(1 << math.ceil(math.log2(n_leaves)))

    if train_sample is None:
        train_sample = n_leaves * 39
        train_sample = min(train_sample, meta['doc_count'])
    elif 0 <= train_sample <= 1:
        train_sample = math.ceil(meta['doc_count'] * train_sample)

    key = ('scann', n_leaves, train_sample)
    index_name = f'scann_leaves-{n_leaves}_train-{train_sample}.scann'
    if key not in self._cache:
        if not os.path.exists(self.index_path/index_name):
            with logger.duration(f'building scann index with {n_leaves} leaves'):
                searcher = scann.ScannBuilder(dvecs, 10, "dot_product") # neighbours=10; doesn't seem to affect the model (?)
                searcher = searcher.tree(num_leaves=n_leaves, num_leaves_to_search=leaves_to_search, training_sample_size=train_sample, quantize_centroids=True)
                searcher = searcher.score_brute_force()
                searcher = searcher.create_pybind()
            with logger.duration('saving scann index'):
                (self.index_path/index_name).mkdir()
                searcher.serialize(str(self.index_path/index_name))
            self._cache[key] = searcher
        else:
            with logger.duration('reading index'):
                self._cache[key] = scann.scann_ops_pybind.load_searcher(dvecs, str(self.index_path/index_name))
    return ScannRetriever(self, self._cache[key], num_results=num_results, leaves_to_search=leaves_to_search, drop_query_vec=drop_query_vec)
FlexIndex.scann_retriever = _scann_retriever



import numpy as np
import torch
import pyterrier as pt
from transformers import AutoConfig
from functools import partialmethod
from .biencoder import BiEncoder, BiQueryEncoder
from .util import Variants
from tqdm import tqdm

def _sbert_encode(self, texts, batch_size=None, prompt=None, normalize_embeddings=False):
    show_progress = False
    if isinstance(texts, tqdm):
        texts.disable = True
        show_progress = True
    texts = list(texts)
    if prompt is not None:
        texts = [prompt + t for t in texts]
    if len(texts) == 0:
        return np.empty(shape=(0, 0))
    return self.model.encode(texts, 
                             batch_size=batch_size or self.batch_size, 
                             show_progress_bar=show_progress,
                             normalize_embeddings=normalize_embeddings
                             )


class SBertBiEncoder(BiEncoder):
    def __init__(self, model_name, batch_size=32, text_field='text', verbose=False, device=None):
        super().__init__(batch_size=batch_size, text_field=text_field, verbose=verbose)
        self.model_name = model_name
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        from sentence_transformers import SentenceTransformer
        self.model = SentenceTransformer(model_name).to(self.device).eval()
        self.config = AutoConfig.from_pretrained(model_name)

    encode_queries = _sbert_encode
    encode_docs = _sbert_encode

    def __repr__(self):
        return f'SBertBiEncoder({repr(self.model_name)})'


class _SBertBiEncoder(SBertBiEncoder, metaclass=Variants):
    VARIANTS: dict = None
    def __init__(self, model_name=None, batch_size=32, text_field='text', verbose=False, device=None):
        super().__init__(model_name or next(iter(self.VARIANTS.values())), batch_size=batch_size, text_field=text_field, verbose=verbose, device=device)

    def __repr__(self):
        inv_variants = {v: k for k, v in self.VARIANTS.items()}
        if self.model_name in inv_variants:
            return f'{self.__class__.__name__}.{inv_variants[self.model_name]}()'
        return super().__repr__()


class Ance(_SBertBiEncoder):
    """Dense encoder for ANCE (Approximate nearest neighbor Negative Contrastive Learning).

    See :class:`~pyterrier_dr.BiEncoder` for usage information.

    .. cite.dblp:: conf/iclr/XiongXLTLBAO21

    .. automethod:: firstp()
    """
    VARIANTS = {
        'firstp': 'sentence-transformers/msmarco-roberta-base-ance-firstp',
    }

class E5(_SBertBiEncoder):
    """Dense encoder for E5 (EmbEddings from bidirEctional Encoder rEpresentations).

    See :class:`~pyterrier_dr.BiEncoder` for usage information.

    .. cite.dblp:: journals/corr/abs-2212-03533

    .. automethod:: base()
    .. automethod:: small()
    .. automethod:: large()
    """

    encode_queries = partialmethod(_sbert_encode, prompt='query: ', normalize_embeddings=True)
    encode_docs = partialmethod(_sbert_encode, prompt='passage: ', normalize_embeddings=True)

    VARIANTS = {
        'base' : 'intfloat/e5-base-v2',
        'small': 'intfloat/e5-small-v2', 
        'large': 'intfloat/e5-large-v2',
    }

class GTR(_SBertBiEncoder):
    """Dense encoder for GTR (Generalizable T5-based dense Retrievers)

    See :class:`~pyterrier_dr.BiEncoder` for usage information.

    .. cite.dblp:: conf/emnlp/Ni0LDAMZLHCY22

    .. automethod:: base()
    .. automethod:: large()
    .. automethod:: xl()
    .. automethod:: xxl()
    """
    VARIANTS = {
        'base': 'sentence-transformers/gtr-t5-base',
        'large': 'sentence-transformers/gtr-t5-large',
        'xl': 'sentence-transformers/gtr-t5-xl',
        'xxl': 'sentence-transformers/gtr-t5-xxl',
    }

class Query2Query(pt.Transformer, metaclass=Variants):
    """Dense query encoder model for query similarity.

    Note that this encoder only provides a :meth:`~pyterrier_dr.BiEncoder.query_encoder` (no document encoder or scorer).

    .. cite:: query2query
        :citation: Bathwal and Samdani. State-of-the-art Query2Query Similarity. 2022.
        :link: https://web.archive.org/web/20220923212754/https://neeva.com/blog/state-of-the-art-query2query-similarity

    .. automethod:: base()
    """
    def __init__(self, model_name=None, batch_size=32, verbose=False, device=None):
        self.model_name = model_name or next(iter(self.VARIANTS.values()))
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        from sentence_transformers import SentenceTransformer
        self.model = SentenceTransformer(self.model_name).to(self.device).eval()
        self.batch_size = batch_size
        self.verbose = verbose

    VARIANTS = {
        'base': 'neeva/query2query',
    }

    encode = _sbert_encode
    transform = BiQueryEncoder.transform
    __repr__ = _SBertBiEncoder.__repr__



from typing import Optional
import numpy as np
import torch
import pyterrier_alpha as pta
import pyterrier as pt
from .. import SimFn, infer_device
from . import FlexIndex
from .np_retr import NumpyScorer


class TorchScorer(NumpyScorer):
    def __init__(self,
        flex_index: FlexIndex,
        torch_vecs: torch.Tensor,
        *,
        num_results: Optional[int] = None
    ):
        self.flex_index = flex_index
        self.torch_vecs = torch_vecs
        self.num_results = num_results
        self.docnos, meta = flex_index.payload(return_dvecs=False)

    def score(self, query_vecs, docids):
        query_vecs = torch.from_numpy(query_vecs).to(self.torch_vecs)
        doc_vecs = self.torch_vecs[docids]
        if self.flex_index.sim_fn == SimFn.cos:
            query_vecs = query_vecs / torch.norm(2, query_vecs, dim=1, keepdims=True)
            doc_vecs = doc_vecs / torch.norm(2, doc_vecs, dim=1, keepdims=True)
            return (query_vecs @ doc_vecs.T).cpu().numpy()
        elif self.flex_index.sim_fn == SimFn.dot:
            return (query_vecs @ doc_vecs.T).cpu().numpy()
        else:
            raise ValueError(f'{self.flex_index.sim_fn} not supported')

    # transform inherited from NumpyScorer


class TorchRetriever(pt.Transformer):
    def __init__(self, 
        flex_index: FlexIndex,
        torch_vecs: torch.Tensor,
        *,
        num_results: int = 1000,
        qbatch: int = 64,
        drop_query_vec: bool = False
    ):
        self.flex_index = flex_index
        self.torch_vecs = torch_vecs
        self.num_results = num_results or 1000
        self.docnos, meta = flex_index.payload(return_dvecs=False)
        self.qbatch = qbatch
        self.drop_query_vec = drop_query_vec

    def transform(self, inp):
        pta.validate.query_frame(inp, extra_columns=['query_vec'])
        inp = inp.reset_index(drop=True)
        query_vecs = np.stack(inp['query_vec'])
        query_vecs = torch.from_numpy(query_vecs).to(self.torch_vecs)

        it = range(0, query_vecs.shape[0], self.qbatch)
        if self.flex_index.verbose:
            it = pt.tqdm(it, desc='TorchRetriever', unit='qbatch')

        result = pta.DataFrameBuilder(['docno', 'docid', 'score', 'rank'])
        for start_idx in it:
            end_idx = start_idx + self.qbatch
            batch = query_vecs[start_idx:end_idx]
            if self.flex_index.sim_fn == SimFn.dot:
                scores = batch @ self.torch_vecs.T
            else:
                raise ValueError(f'{self.flex_index.sim_fn} not supported')
            if scores.shape[1] > self.num_results:
                scores, docids = scores.topk(self.num_results, dim=1)
            else:
                docids = scores.argsort(descending=True, dim=1)
                scores = torch.gather(scores, dim=1, index=docids)
            for s, d in zip(scores.cpu().numpy(), docids.cpu().numpy()):
                result.extend({
                    'docno': self.docnos[d],
                    'docid': d,
                    'score': s,
                    'rank': np.arange(s.shape[0]),
                })

        if self.drop_query_vec:
            inp = inp.drop(columns='query_vec')
        return result.to_df(inp)


def _torch_vecs(self, *, device: Optional[str] = None, fp16: bool = False) -> torch.Tensor:
    """Return the indexed vectors as a pytorch tensor.

    .. caution::
        This method loads the entire index into memory on the provided device. If the index is too large to fit in memory,
        consider using a different method that does not fully load the index into memory, like :meth:`np_vecs` or
        :meth:`get_corpus_iter`.

    Args:
        device: The device to use for the tensor. If not provided, the default device is used (cuda if available, otherwise cpu).
        fp16: Whether to use half precision (fp16) for the tensor.

    Returns:
        :class:`torch.Tensor`: The indexed vectors as a torch tensor.
    """
    device = infer_device(device)
    key = ('torch_vecs', device, fp16)
    if key not in self._cache:
        dvecs, meta = self.payload(return_docnos=False)
        data = torch.frombuffer(dvecs, dtype=torch.float32).reshape(*dvecs.shape)
        if fp16:
            data = data.half()
        self._cache[key] = data.to(device)
    return self._cache[key]
FlexIndex.torch_vecs = _torch_vecs


def _torch_scorer(self, *, num_results: Optional[int] = None, device: Optional[str] = None, fp16: bool = False):
    """Return a scorer that uses pytorch to score (re-rank) results using indexed vectors.

    The returned :class:`pyterrier.Transformer` expects a DataFrame with columns ``qid``, ``query_vec`` and ``docno``.
    (If an internal ``docid`` column is provided, this will be used to speed up vector lookups.)

    .. caution::
        This method loads the entire index into memory on the provided device. If the index is too large to fit in memory,
        consider using a different scorer that does not fully load the index into memory, like :meth:`np_scorer`.

    Args:
        num_results: The number of results to return per query. If not provided, all resuls from the original fram are returned.
        device: The device to use for scoring. If not provided, the default device is used (cuda if available, otherwise cpu).
        fp16: Whether to use half precision (fp16) for scoring.

    Returns:
        :class:`~pyterrier.Transformer`: A transformer that scores query vectors with pytorch.
    """
    return TorchScorer(self, self.torch_vecs(device=device, fp16=fp16), num_results=num_results)
FlexIndex.torch_scorer = _torch_scorer


def _torch_retriever(self,
    *,
    num_results: int = 1000,
    device: Optional[str] = None,
    fp16: bool = False,
    qbatch: int = 64,
    drop_query_vec: bool = False
):
    """Return a retriever that uses pytorch to perform brute-force retrieval results using the indexed vectors.

    The returned :class:`pyterrier.Transformer` expects a DataFrame with columns ``qid``, ``query_vec``.

    .. caution::
        This method loads the entire index into memory on the provided device. If the index is too large to fit in memory,
        consider using a different retriever that does not fully load the index into memory, like :meth:`np_retriever`.

    Args:
        num_results: The number of results to return per query.
        device: The device to use for scoring. If not provided, the default device is used (cuda if available, otherwise cpu).
        fp16: Whether to use half precision (fp16) for scoring.
        qbatch: The number of queries to score in each batch.
        drop_query_vec: Whether to drop the query vector from the output.

    Returns:
        :class:`~pyterrier.Transformer`: A transformer that retrieves using pytorch.
    """
    return TorchRetriever(self, self.torch_vecs(device=device, fp16=fp16), num_results=num_results, qbatch=qbatch, drop_query_vec=drop_query_vec)
FlexIndex.torch_retriever = _torch_retriever
